import os
import shutil
import sys
from time import time
import pandas as pd
from subprocess import run
from obspy import read_inventory
from tqdm import tqdm
import glob
import SeisMonitor.utils as sut

CORE_NLLOC = os.path.join(os.path.dirname(__file__),"core")
NLLOC_path = os.path.join(CORE_NLLOC,"NonLinLoc-main")
src_path = os.path.join(NLLOC_path,"src")
bin_path = os.path.join(src_path,"bin")
vel2grid_exe_path = os.path.join(bin_path,"Vel2Grid")
grid2time_exe_path = os.path.join(bin_path,"Grid2Time")
nll_exe_path = os.path.join(bin_path,"NLLoc")

def run_nlloc(p_control_file_path,
            s_control_file_path):

    sut.printlog("info","NLLoc:Vel2Grid", "Running")
    vel2grid = os.system(f"{vel2grid_exe_path} {p_control_file_path} > /dev/null")
    sut.printlog("info","NLLoc:Grid2Time:P", "Running")
    grid2time = os.system(f"{grid2time_exe_path} {p_control_file_path} > /dev/null")
    sut.printlog("info","NLLoc:Grid2Time:S", "Running")
    grid2time = os.system(f"{grid2time_exe_path} {s_control_file_path} > /dev/null")
    sut.printlog("info","NLLoc:NLLoc", "Running")
    grid2time = os.system(f"{nll_exe_path} {s_control_file_path} > /dev/null")

def apt_install(pkgs):
    cmd = ['pkexec', 'apt-get', 'install', '-y'] + pkgs
    print('Running command: {}'.format(' '.join(cmd)))
    result = run(
        cmd,
        stdout=sys.stdout,
        stderr=sys.stderr,
        encoding='utf8',
        env={**os.environ, 'DEBIAN_FRONTEND': 'noninteractive'}
    )
    result.check_returncode()

def write_pref_origin_removing_phaselocinfo(catalog):
    events = []
    for ev in catalog:
        pref_origin = ev.preferred_origin()

        # replace none by NN
        for pick in ev.picks:
            pick.waveform_id.network_code= "NN"
            pick.waveform_id.location_code= "NN"
            pick.waveform_id.channel_code= "NNN"
            pick.evaluation_mode = "automatic"


        new_arrivals = []
        for arrival in pref_origin.arrivals:
            arrival.azimuth = None
            arrival.distance = None
            arrival.takeoff_angle=None
            arrival.time_residual=None
            arrival.time_weight=None
            new_arrivals.append(arrival)

        for i,origin in enumerate(ev.origins):
            if origin.resource_id.id == ev.preferred_origin_id:
                ev.origins[i].arrivals = new_arrivals 
            # else:
            #     continueagency_id=self.agency

        del ev.origins
        ev.origins = [pref_origin]
        events.append(ev)
    catalog.events = events
    return catalog

def download_nlloc(forced=False):
    name = "nll.zip"
    zip_path = os.path.join(CORE_NLLOC,name)
    cache_path = os.path.join(src_path,"CMakeCache.txt")
    # "https://github.com/alomax/NonLinLoc/archive/refs/heads/main.zip"

    if not forced:
        if os.path.isfile(nll_exe_path):
            return True
        else:
            pass

    isfile = sut.isfile(zip_path)
    if not isfile:
        os.system(f"wget https://github.com/alomax/NonLinLoc/archive/refs/heads/main.zip -O {zip_path}")

    if not os.path.isdir(NLLOC_path):
        os.system(f"unzip {zip_path} -d {CORE_NLLOC}")

    if os.path.isdir(bin_path):
        os.rmdir(bin_path)
        os.makedirs(bin_path)

    if os.path.isdir(cache_path):
        os.rmdir(cache_path)

    try:
        apt_install(["cmake"])
    except:
        sut.printlog("warning","Install Cmake","Could not install Cmake")

    try:
        os.system(f"cd {src_path} && cmake . && make")
    except:
        raise Exception("Could not compile NLLoc")

def write_1d_vel_model(vel_path,out,
                        compute_vs=True,
                        vp_vs_ratio=1.78):
    df = pd.read_csv(vel_path)

    vm = open(out, 'w')
    msg ="# model layers (LAYER depth, Vp_top, Vp_grad, Vs_top, Vs_grad, p_top, p_grad)\n"
    vm.write(msg)
    for i,row in df.iterrows():
        if compute_vs:
            vs = row.vp/vp_vs_ratio
        else:
            vs = row.vs

        if i == len(df)-1:
            enter = ""
        else:
            enter = "\n"

        msg = f"LAYER    {row.depth:<6.2f}    {row.vp:<.2f}    0.00    {vs:<.2f}    0.00    {row.rho:<.2f}    0.00{enter}"
        vm.write(msg)
    vm.close()

def resp2df(resp):
    """
    Parameters:
    -----------
    resp: str
        RESP filepath

    Returns: DataFrame
        Dataframe with the next columns
        network,station,latitude,longitude,elevation
    """
    networks = []
    stations = []
    longitudes = []
    latitudes = []
    elevations = []
    inv = read_inventory(resp)
    for net in inv:
        for sta in net:
            latitudes.append(sta.latitude)
            longitudes.append(sta.longitude)
            elevations.append(sta.elevation)
            stations.append(sta.code)
            networks.append(net.code)

    df = {"network":networks,"station":stations,
        "latitude":latitudes,"longitude":longitudes,
        "elevation":elevations}
    df = pd.DataFrame(df)
    return df 

def write_station_file(sta_path,out):
    df = resp2df(sta_path)

    vs = open(out, 'w')
    msg = "# GTSRCE label LATLON latSrce longSrce zSrce elev\n"
    vs.write(msg)
    for i,row in df.iterrows():
        elv = row.elevation/1e3
        if i == len(df)-1:
            enter = ""
        else:
            enter = "\n"

        msg = f"GTSRCE  {row.station:<5}  LATLON  {row.latitude:<7.3f}  {row.longitude:<8.3f}  0.000  {elv:<7.3f}{enter}"
        vs.write(msg)
    vs.close()

def join_args(args:str):
    return " ".join(list(map(lambda x: str(x),args)))
    
class GenericControlStatement():
    def __init__(self,
                trans:list,
                control:list=[1,54321]):
        self.control = join_args(control)
        self.trans = join_args(trans)
        sut.validate(self.__init__, locals())

    def get_msg(self):
        msg = (
            "#__________________START GENERIC CONTROL STATEMENTS\n\n"
            "CONTROL {}\n",
            "TRANS {}\n",
            "#__________________END\n"

            )
        msg = "".join(msg)
        msg = msg.format(self.control,
                    self.trans)
        sut.printlog("debug","NLLoc:GenericControlStatement",
                    "Message received")
        return msg

class Vel2Grid():
    def __init__(self,
                vel_path:str,
                grid_folder_out:str,
                grid:list,
                p_phase:str="P",
                s_phase:str="S") -> None:
        self.vel_path = vel_path
        self.grid = join_args(grid)
        self.grid_folder_out = grid_folder_out
        self.p_phase = p_phase
        self.s_phase = s_phase
        sut.validate(self.__init__, locals())

    def get_msg(self):
        msg = (
            "#__________________START VEL2GRID STATEMENTS\n\n"
            "VGOUT {}\n",
            "VGTYPE {}\n",
            "VGTYPE {}\n",
            "VGGRID {}\n",
            "INCLUDE {}\n",
            "#__________________END\n"

            )
        msg = "".join(msg)
        msg = msg.format(self.grid_folder_out,
                        self.p_phase,
                        self.s_phase,
                        self.grid,
                        self.vel_path)

        path = os.path.dirname(self.grid_folder_out)
        sut.isfile(path)

        sut.printlog("debug","NLLoc:Vel2grid",
                    "Message received")
        return msg

class Grid2Time():
    def __init__(self,
        station_path:str,
        grid_folder_out:str,
        time_folder_out:str,
        phase:str="P",
        mode:list=["GRID3D","ANGLES_YES"],
        plfd:list=[1e-3,0]) -> None:

        self.station_path = station_path 
        self.grid_folder_out = grid_folder_out 
        self.time_folder_out = time_folder_out 
        self.phase = phase
        self.mode = join_args(mode) 
        self.plfd = join_args(plfd) 
        sut.validate(self.__init__, locals())

    def get_msg(self):
        msg = (
            "#__________________START GRID2TIME STATEMENTS\n\n",
            "GTFILES {} {} {}\n",
            "GTMODE {}\n",
            "INCLUDE {}\n",
            "GT_PLFD {}\n"
            "#__________________END\n"
            )
        msg = "".join(msg)
        msg = msg.format(self.grid_folder_out,self.time_folder_out,self.phase,
                self.mode,
                self.station_path,
                self.plfd 
                )

        path = os.path.dirname(self.time_folder_out)
        sut.isfile(path)

        sut.printlog("debug","NLLoc:Grid2Time",
                    "Message received")
        return msg

class Time2Loc():
    def __init__(self,
        catalog:list,
        grid:list,
        time_folder_out:str,
        loc_folder_out:str,
        meth:list = ["GAU_ANALYTIC",9999,4,-1,-1,1.78,6],
        # meth:list = ["EDT_OT_WT",9999,4,-1,-1,1.78,6],
        search:list = ["OCT",37,58,7,1e-2,int(1e5),int(1e4)],
        sig:str = "SeisMonitor",
        com:str = "Comment",
        gau:list = [0.2,0.0],
        gau2:list = [0.05,0.05,2.0],
        p_phaseid:list = ["P","P","p","PN","PG","Pn","Pg"],
        s_phaseid:list = ["S","S","s","SN","SG","Sn","Sg"],
        qual2err:list = [0.1, 0.5, 1.0, 2.0, 99999.9],
        phstat:list = [9999.0,-1,9999.0,1.0,1.0,9999.0,-9999.0,9999.0],
        angles:list = ["ANGLES_YES",5],
        hypout:list = ["SAVE_NLLOC_ALL","SAVE_NLLOC_SUM",
                    "SAVE_HYPO71_SUM"],
        mag:list = ["ML_HB",1.0,1.110,0.00189]
        ):
        self.catalog = join_args(catalog)
        self.time_folder_out = time_folder_out
        self.loc_folder_out = loc_folder_out
        self.grid = join_args(grid)
        self.meth = join_args(meth)
        self.search = join_args(search)
        self.sig = sig
        self.com = com
        self.gau = join_args(gau)
        self.gau2 = join_args(gau2)
        self.p_phaseid = join_args(p_phaseid)
        self.s_phaseid = join_args(s_phaseid)
        self.qual2err = join_args(qual2err)
        self.phstat = join_args(phstat)
        self.angles = join_args(angles)
        self.hypout = join_args(hypout)
        self.mag = join_args(mag)
        sut.validate(self.__init__, locals())

    def get_msg(self):
        msg = (
            "#__________________START NLDIFFLOC STATEMENTS\n\n"
            "LOCSIG {}\n",
            "LOCCOM {}\n",
            "LOCFILES {} {} {}\n",
            "LOCHYPOUT {}\n",
            "LOCSEARCH {}\n",
            "LOCGRID {}\n",
            "LOCMETH {}\n",
            "LOCGAU {}\n",
            "LOCGAU2 {}\n",
            "LOCPHASEID {}\n",
            "LOCPHASEID {}\n",
            "LOCQUAL2ERR {}\n",
            "LOCPHSTAT {}\n",
            "LOCANGLES {}\n",
            "LOCMAG {}\n",
            "#__________________END\n")
            
        msg = "".join(msg)
        msg = msg.format(
                        self.sig,
                        self.com,
                        self.catalog,self.time_folder_out,self.loc_folder_out,
                        self.hypout,
                        self.search,
                        self.grid,
                        self.meth,
                        self.gau,
                        self.gau2,
                        self.p_phaseid,
                        self.s_phaseid,
                        self.qual2err,
                        self.phstat,
                        self.angles,
                        self.mag)

        path = os.path.dirname(self.loc_folder_out)
        sut.isfile(path)

        sut.printlog("debug","NLLoc:Time2Loc",
                    "Message received")
        return msg

class NLLocControlFile():
    def __init__(self,
                generic_control:GenericControlStatement,
                vel2grid:Vel2Grid,
                grid2time:Grid2Time,
                time2loc:Time2Loc):
        self.generic_control = generic_control
        self.vel2grid = vel2grid
        self.grid2time = grid2time
        self.time2loc = time2loc
        sut.validate(self.__init__, locals())

    def _validate_args(self,key:str,input:str,output:str):
        sut.printlog("debug","NLLoc:validate_control_file_args",
                    "validating control file args")

        if os.path.isfile(input):
            sut.printlog("debug","NLLoc:validate_control_file_args",
                f"{key} is ok.")
        else:
            raise Exception(f"NLLoc: {key} path doesn't exist. Check:{input}")

        sut.isfile(output)

        sut.printlog("debug","NLLoc:validate_control_file_args",
                    "control file args are ok")

    def get_msg(self,
            vel2grid:bool=True,
            grid2time:bool=True,
            time2loc:bool=True):

        msg = self.generic_control.get_msg()

        if vel2grid:
            msg += self.vel2grid.get_msg()
            input = self.vel2grid.vel_path
            output = self.vel2grid.grid_folder_out
            self._validate_args("vel_path",input,output)
        if grid2time:
            input = self.grid2time.station_path
            output = self.grid2time.time_folder_out
            self._validate_args("station_path",input,output)
            msg += self.grid2time.get_msg()
        if time2loc:
            input = self.time2loc.catalog.split(" ")[0]
            output = self.time2loc.loc_folder_out
            self._validate_args("catalog_path",input,output)
            msg += self.time2loc.get_msg()

        return msg

    def write(self,
            out:str,
            vel2grid:bool=True,
            grid2time:bool=True,
            time2loc:bool=True):

        sut.printlog("info","NLLoc:write_control_file",
                    "Running")
        sut.isfile(out)

        msg = self.get_msg(vel2grid,grid2time,time2loc)

        control_file_msg = open(out,"w")
        control_file_msg.write(msg)
        control_file_msg.close()
        sut.printlog("info","NLLoc:write_control_file",
                    f"Finished. Control file: {out} ")

if __name__=="__main__":
    catalog = "/home/emmanuel/EDCT/SeisMonitor/SeisMonitor/monitor/locator/nlloc/test3/select.out"
    station_path = "/home/emmanuel/EDCT/SeisMonitor/SeisMonitor/monitor/locator/nlloc/test3/station.dat"
    vel_path = "/home/emmanuel/EDCT/SeisMonitor/SeisMonitor/monitor/locator/nlloc/test3/model.dat"
    grid_folder_out = "/home/emmanuel/EDCT/SeisMonitor/SeisMonitor/monitor/locator/nlloc/test3/model/layer"
    time_folder_out = "/home/emmanuel/EDCT/SeisMonitor/SeisMonitor/monitor/locator/nlloc/test3/time/layer"
    loc_folder_out = "/home/emmanuel/EDCT/SeisMonitor/SeisMonitor/monitor/locator/nlloc/test3/loc/SeisMonitor"
    control_file_out = "/home/emmanuel/EDCT/SeisMonitor/SeisMonitor/monitor/locator/nlloc/test3/test.in"

    gen_control = GenericControlStatement(trans=["SIMPLE",
                                             5.0,-73.0,0.0])
    vel2grid = Vel2Grid(vel_path,
                        grid_folder_out,
                        grid = [2,583,70,
                                -891.0,-891.0,-5.0,
                                3.0, 3.0, 3.0,
                                "SLOW_LEN"],
                        phase = "P")
    grid2time = Grid2Time(station_path,
                            grid_folder_out,
                            time_folder_out)
    time2loc = Time2Loc(catalog=[catalog,"SEISAN"],
                        grid = [374,583,70,
                                -891.0,-891.0,-5.0,
                                3.0,3.0,3.0,
                                "PROB_DENSITY","SAVE"],
                        time_folder_out=time_folder_out,
                        loc_folder_out=loc_folder_out)
    nlloc = NLLocObj(gen_control,vel2grid,
                    grid2time,time2loc)
    nlloc.write_control_file(control_file_out)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Nov 12 20:00:00 2020
@author: Emmanuel_Castillo
last update: 14-11-2020 
"""

import os
import json
import time
import logging
import datetime as dt
from SeisMonitor.utils import printlog
from obspy.core.stream import Stream
from obspy.clients.fdsn.mass_downloader import Restrictions
from obspy.core.inventory.inventory import (Inventory,read_inventory)
from obspy.clients.fdsn.mass_downloader.domain import RectangularDomain

class DownloadRestrictions():
	def __init__(self,mseed_storage, 
                chunklength_in_sec=None,
                threshold= 60,
                overlap_in_sec=0,
                picker_args={}, # {"batch_size":10,"overlap":0.3,"lenght":60}
                groupby='{network}.{station}.{channel}',
                n_processor=None) -> None:
		self.mseed_storage = mseed_storage
		self.chunklength_in_sec = chunklength_in_sec
		self.threshold = threshold
		self.overlap_in_sec = overlap_in_sec
		self.picker_args = picker_args
		self.groupby = groupby
		self.n_processor=n_processor
		

def sanitize_provider_times(providers):
    provider_times = [(provider.waveform_restrictions.starttime,
                        provider.waveform_restrictions.endtime)\
                        for provider in providers]
    if provider_times.count(provider_times[0]) == len(provider_times):
        return  providers
    else:
        raise Exception("Providers must have the same interval time")

def get_max_allowed_batch_size(data_length,segment_length,overlap):
    """
    data_length : length of the data in seconds
    segment_length: length of each batch segment in seconds 
    overlap: overlap (0-1)
    """
    max_batch_size = (data_length-(overlap*segment_length)) / (segment_length*(1-overlap))
    max_batch_size = int(max_batch_size)

    if max_batch_size == 0:
        max_batch_size = 1

    return max_batch_size

def write_stream(st,mseed_storage, 
				threshold=None, 
				picker_args={}, # {"batch_size":10,"overlap":0.3,"lenght":60}
				ppc_and_comment = [False,""]):
	"""
	Write a stream in a specific storage given by mseed_storage

	Parameters:
	-----------
	st: obspy.Stream object
		Stream that will be written.
	ppc_restrictions: PreprocRestrictions object
		Restrictions to preprocess a stream.
	mseed_storage:
		Where to store the waveform files.
		The parameter should name the corresponding keys 
		of the stats object,
		e.g. '{network}.{station}.{location}.{channel}__{starttime}__{endtime}
	threshold: int
		data length in seconds
	picker_args : dict
		keys: batch_size,overlap,length
		It's used to know if the stream can be downloaded according to the 
		picker keys. If the the segments given by the length of the stream
		and overlap parammeter are less than batch_size, then no download the stream.

	Returns:
		write one stream
	"""
	

	ppc,comment = ppc_and_comment
	tr = st[0]

	mseed_filename = get_mseed_filename(_str=mseed_storage, 
									tr=tr,
									ppc=ppc)

	if threshold != None:
		length = abs(tr.stats.endtime - tr.stats.starttime)
		if length < threshold:
			comment = f"length:{length} < threshold:{threshold}"
			# logger =logging.getLogger('Downloaded: False') 
			# logger.info(f'{mseed_filename}  {comment}')
			printlog("info",'Downloader: False',
					f'{mseed_filename}  {comment}')
			download = False
		else:
			download = True

	if picker_args:
		overlap,batch_size,segment_length = picker_args["overlap"],picker_args["batch_size"],picker_args["length"]
		data_length = abs(tr.stats.endtime - tr.stats.starttime)
		max_batch_size = get_max_allowed_batch_size(data_length,segment_length,overlap)
		
		if max_batch_size < batch_size:
			comment = (f"This mseed only can be used with {max_batch_size} batchs. "+\
					f"In order to download the data, the batch size must be >= {batch_size}."+\
					f" Modify this condition changing 'picker_args':{picker_args}' parameter.")
			# logger =logging.getLogger('Downloaded: False') 
			# logger.info(f'{mseed_filename}  {comment}')
			printlog("info",'Downloader: False',
					f'{mseed_filename}  {comment}')
			download = False
		else:
			download = True
	if (threshold == None) or (bool(picker_args) == False):
		download = True

	if os.path.isfile(mseed_filename) == False:

		if download == False:
			return
		else:
			mseed_dir = os.path.dirname(mseed_filename)
			if os.path.isdir(mseed_dir) == False:
				os.makedirs(mseed_dir)
			else:
				pass
			st.write(mseed_filename,format="MSEED")
			# logger =logging.getLogger('Downloaded: True')
			# logger.info(f'{mseed_filename}  {comment}') 

			printlog("info",'Downloader: True',
					f'{mseed_filename}  {comment}')

	else:
		# logger =logging.getLogger('Downloaded: Exist')
		# logger.info(f'{mseed_filename}  {comment}') 
		printlog("info",'Downloader: Exist',
					f'{mseed_filename}  {comment}')
	del tr; del st


def get_mseed_filename(_str, tr,ppc=False):
	"""
	It generates the path of the file.

	_str : str
		string with the wildcards
		possible wildcards: network,station,location,channel,starttime,endtime,year,month,day,julday
	tr: Trace
		trace to take the information
	ppc: bool
		True if it was preprocesed. If true .pcc is added in the path.
	"""
	strftime = "%Y%m%dT%H%M%SZ"
	network=tr.stats.network 
	station=tr.stats.station
	location=tr.stats.location 
	channel=tr.stats.channel
	starttime=tr.stats.starttime 
	_starttime=starttime.strftime(strftime) 
	endtime=tr.stats.endtime
	_endtime=endtime.strftime(strftime)
	year = starttime.year
	month = starttime.month
	day = endtime.day
	julday = endtime.julday
	path = _str.format(
                    network=network, station=station, location=location,
                    channel=channel, year=year, month=month, 
                    day=day, julday=julday,starttime=_starttime,endtime=_endtime)
	if ppc:
		path = path + ".ppc"

	if not isinstance(path, (str, bytes)):
		raise TypeError("'%s' is not a filepath." % str(path))
	return path

def get_chunktimes(starttime,endtime,chunklength_in_sec, overlap_in_sec=0):
	"""
	Make a list that contains the chunktimes according to 
	chunklength_in_sec and overlap_in_sec parameters.

	Parameters:
	-----------
	starttime: obspy.UTCDateTime object
		Start time
	endtime: obspy.UTCDateTime object
		End time
	chunklength_in_sec: None or int
		The length of one chunk in seconds. 
		The time between starttime and endtime will be divided 
		into segments of chunklength_in_sec seconds.
	overlap_in_sec: None or int
		For more than one chunk, each segment will have overlapping seconds

	Returns:
	--------
	times: list
		List of tuples, each tuple has startime and endtime of one chunk.
	"""

	if chunklength_in_sec == 0:
		raise Exception("chunklength_in_sec must be different than 0")
	elif chunklength_in_sec == None:
		return [(starttime,endtime)]

	if overlap_in_sec == None:
		overlap_in_sec = 0

	deltat = starttime
	dtt = dt.timedelta(seconds=chunklength_in_sec)
	overlap_dt = dt.timedelta(seconds=overlap_in_sec)

	times = []
	while deltat < endtime:
		# chunklength can't be greater than (endtime-startime)
		if deltat + dtt > endtime:
			break
		else:
			times.append((deltat,deltat+dtt))
			deltat += dtt - overlap_dt

	if deltat < endtime:	
		times.append((deltat,endtime))
	# print(times)
	return times

def get_st_according2preference(st,location_list,channel_list):
	"""
	Suppose that your preference_type is "location" and your 
	location_list is ["00","20","10"], then this function first
	filter the stream according to location and returns
	a  new stream only with location "00", if no exist "00" will 
	continue with the next preference "20", and otherwise, "10". 
	After that, it is going to take new stream and it will go to 
	filter according to channel_list preference if the new stream 
	has more than one channel type ("HH or BH). 

	Parameters:
	-----------
	st: stream object
		stream 
	location_list: list
		locations in order of the preference ["00","20","10"]
	channel_list: list
		channels in order of the preference ["HH","BH"]

	results:
	--------
	new_st : stream object
		stream according to the preference
	"""
	logger =logging.getLogger(f'Downloader: preference')
	if len(st) >= 1000:
		stats = st[0].stats
		# logger.error(f"{stats.network}-{stats.station}: "+
		# f"No filter by preference. gaps: {len(st)}  ")

		printlog("error",'Downloader: preference',
				f"{stats.network}-{stats.station}"+
				f"No filter by preference. gaps: {len(st)}")
		return st

	preference = list(map(lambda x: (x[2],x[3]),st._get_common_channels_info().keys() ))

	if not location_list:
		stats = st[0].stats
		new_st = st
	else:
		locations = list(map(lambda x: x[2],st._get_common_channels_info().keys() ))
	
		index = 0
		loc_pref = None
		# print(len(location_list))
		while index < len(location_list):
			loc_pref = location_list[index]
			if loc_pref in locations:
				index = len(location_list)
			else:
				loc_pref = None
				index += 1

		if loc_pref == None:
			return st

		stats = st[0].stats
		new_st = st.select(network=stats.network, station = stats.station,
							location=loc_pref)

	if not channel_list:
		pass
	else:
		## If the same location has two differents sensor
		## example : 00.HH? or 00.BH?, then choose 
		common_channels = new_st._get_common_channels_info().keys()

		common_channels = list(common_channels)
		common_channels = list(map(lambda x: x[3][:2],common_channels))
		index = 0
		cha_pref = None
		while index < len(channel_list):
			cha_pref = channel_list[index]
			if cha_pref in common_channels:
				index = len(channel_list)
			else:
				cha_pref = None
				index += 1

		if cha_pref == None:
			return st
		else:
			cha_pref = f'{cha_pref}?'
		

		new_st = new_st.select(network=stats.network, station = stats.station,
								location=loc_pref, channel=cha_pref)

	common_channels = list(map(lambda x: (x[2],x[3]),new_st._get_common_channels_info().keys()))

	printlog("debug",f"Downloader: {stats.network}-{stats.station}: ",
				f"available:{preference},"+
				f" selected {common_channels} according to preference")
	# logger.info(f"{stats.network}-{stats.station}: "+
	# 	f"available:{preference},"+
	# 	f" selected {common_channels} according to preference")
	return new_st

def get_filenames(mseed, filter_net=[],
				filter_sta=[],filter_cha=[]):
	"""
	Parameters:
	-----------
	mseed: str
		path of the folder where you need to get all mseed filenames.
	filter_net: list
		avoid mseed if its filename contain one network specified in
		filter net
	filter_sta: list
		avoid mseed if its filename contain one station specified in
		filter net
	filter_cha: list
		avoid mseed if its filename contain one channel specified in
		filter net

	Returns:
	--------
	list
	Get all filenames contained in mseed folder.
	"""

	if not isinstance(filter_net, list):
		raise Exception("filter_net is a list")
	elif not isinstance(filter_sta, list):
		raise Exception("filter_sta is a list")
	elif not isinstance(filter_cha, list):
		raise Exception("filter_cha is a list")

	filenames = []
	file_list = [ev for ev in os.listdir(mseed) \
				if ev.split('/')[-1].split('.')[-1].lower() == 'mseed']
	for onefile in file_list:
		filename = onefile.split('/')[-1]
		info = filename.split('__')[0]
		net,sta,loc,cha = info.split('.')
		if net in filter_net:
			pass
		if sta in filter_sta:
			pass
		if len(filter_cha) != 0:
			if cha not in filter_cha:
				pass
		else:
			filenames.append(filename)
	return filenames

def get_all_sdswaveforms(client,**kwargs):
	"""
	client: Client from obspy
		Client to get_waveforms
	**kwargs: 
		network,station,location,channel

	"""
	myargs = {}
	for key,value in kwargs.items():
		if key in ("network","station","location","channel"):
			myvalue = value.split(",")
		else: 
			myvalue = value
		myargs[str(key)] = myvalue

	st = Stream()
	for net in myargs["network"]:
		for sta in myargs["station"]:
			for loc in myargs["location"]:
				for cha in myargs["channel"]:
					try:
						myst = client.get_waveforms(net,sta,loc,cha,
								myargs['starttime'],myargs['endtime'])

						if len(myst) != 0:
							st += myst

					except Exception as e:
						# logger = logging.getLogger('get waveform')
						# logger.warning(f'not obtained: {net}.{sta}.{loc}.{cha}.
						# 		{myargs['starttime']}.{myargs['endtime']}:{e}') 
						# print(f"not obtained: {net}.{sta}.{loc}.{cha}."+\
						#	   f"{myargs['starttime']}.{myargs['endtime']}:{e}")
						seedname = f"{net}.{sta}.{loc}.{cha}.{myargs['starttime']}.{myargs['endtime']}"
						printlog("error",seedname,f"{e}")

	return st

def select_inventory(inv,network,station,location,
                    channel,starttime,endtime):
    """
    Filter inventory according to parameters
    Parameters
    ----------
    inv:Inventory
        Inventory that will be filtered
    network: str
        Select one or more network codes. 
        Multiple codes are comma-separated (e.g. "IU,TA"). 
        Wildcards are allowed.
    station: str
        Select one or more SEED station codes. 
        Multiple codes are comma-separated (e.g. "ANMO,PFO"). 
        Wildcards are allowed.
    location: str
        Select one or more SEED location identifiers. 
        Multiple identifiers are comma-separated (e.g. "00,01"). 
        Wildcards are allowed.
    channel: str
        Select one or more SEED channel codes. 
        Multiple codes are comma-separated (e.g. "BHZ,HHZ").
    starttime: obspy.UTCDateTime
        Limit results to time series samples on or 
        after the specified start time.
    endtime: obspy.UTCDateTime
        Limit results to time series samples on or 
        before the specified end time.
    Returns:
    --------
    inventory: Inventory
        Filtered Inventory
    """
    networks = network.split(',')
    stations = station.split(',')
    locations = location.split(',')
    channels = channel.split(',')

    cha0 = channels[0]
    inventory = inv.select(network=networks[0],
                            station=stations[0],
                            location=locations[0],
                            channel=cha0,
                            starttime=starttime,
                            endtime=endtime)
    for net in networks:
        for sta in stations:
            for loc in locations:
                for cha in channels:
                    one_inv = inv.select(network=net,station=sta,
                                            location=loc,channel=cha,
                                            starttime=starttime,
                                            endtime=endtime)
                    inventory = inventory.__add__(one_inv) 
    return inventory

def get_client_waveforms(client,bulk,
						waveform_restrictions,
						processing):
	strftime = "%Y%m%dT%H%M%SZ"
	net,sta,loc,cha,starttime,endtime = bulk
	loc_preference = waveform_restrictions.location_preferences
	cha_preference = waveform_restrictions.channel_preferences

	why = "-".join((net,sta,loc,cha,starttime.strftime(strftime),endtime.strftime(strftime)))
	try:
		st = client.get_waveforms(net,sta,loc,cha,starttime,endtime)
	except Exception as e:
		printlog("info","Downloader: False",why+"->"+str(e))
		st = Stream()
		ppc = False
		comment = ""
		return st,ppc,comment
	# print(st)
	if len(st)==0:
		printlog("warning","Downloader: False","0 Trace(s) in Stream: "+why)
		ppc = False
		comment = ""
		return st,ppc,comment

	st= get_st_according2preference(st,
								loc_preference,
								cha_preference)	
	if processing == None:
		ppc = False
		comment = ""
	else:
		st,ppc,comment = processing.run(st)
	return st,ppc,comment

def write_client_waveforms(client,bulk,
						waveform_restrictions,
						download_restrictions,
						processing):
	st,ppc,comment = get_client_waveforms(client,bulk,
						waveform_restrictions,
						processing)
	groupby = download_restrictions.groupby
	st_dict = st._groupby(groupby)
	for st in st_dict.values():
		write_stream(st,
					download_restrictions.mseed_storage,
					download_restrictions.threshold,
					download_restrictions.picker_args,
					[ppc,comment])
	# del st;st_dict;del client;del bulk;del waveform_restrictions;del processing
	# return st_dict

def get_merged_inv_and_json(providers):
	json_info = {}
	stations_outside_domains = []
	inventory = Inventory()
	updated_providers = []
	for provider in providers:
		client = provider.client
		restrictions = provider.waveform_restrictions
		if provider.xml != None:
			# print("with xml")
			# print(restrictions.__dict__)
			try:
				inv = read_inventory(provider.xml)
				inv = select_inventory(inv=inv,network=restrictions.network,
											station=restrictions.station,
											location=restrictions.location,
											channel=restrictions.channel,
											starttime=restrictions.starttime, 
											endtime=restrictions.endtime)
			except:
				printlog("error","Inventory",f"No get_stations with {restrictions.__dict__}")
				inv = Inventory()
		else:
			# print("no xml")
			# print(restrictions.__dict__)
			try:
				inv=  client.get_stations(network=restrictions.network,
															station=restrictions.station,
															location=restrictions.location,
															channel=restrictions.channel,
															starttime=restrictions.starttime, 
															endtime=restrictions.endtime, 
															level='channel')
			except:
				printlog("error","Inventory",f"No get_stations with {restrictions.__dict__}")
				inv = Inventory()
				

		inv,jinf,sod = get_inv_and_json(inv, 
									restrictions.filter_networks, 
									restrictions.filter_stations,
									restrictions.filter_domain)

		bulk_info = list(set([(net.code,sta.code) for net in inv.networks for sta in net.stations ]))

		provider.waveform_restrictions.bulk_info = bulk_info

		inventory = inventory.__add__(inv) 
		json_info.update(jinf)
		stations_outside_domains.append(sod)
		updated_providers.append(provider)

	stations_outside_domains = [x for xs in stations_outside_domains for x in xs]
	return inventory,json_info,updated_providers,stations_outside_domains

def inside_the_polygon(p,pol_points):
    """
    Parameters:
    -----------
    p: tuple
        Point of the event. (lon,lat)
    pol_points: list of tuples
        Each tuple indicates one polygon point (lon,lat).
    Returns: 
    --------
    True inside 
    """
    V = pol_points

    cn = 0  
    V = tuple(V[:])+(V[0],)
    for i in range(len(V)-1): 
        if ((V[i][1] <= p[1] and V[i+1][1] > p[1])   
            or (V[i][1] > p[1] and V[i+1][1] <= p[1])): 
            vt = (p[1] - V[i][1]) / float(V[i+1][1] - V[i][1])
            if p[0] < V[i][0] + vt * (V[i+1][0] - V[i][0]): 
                cn += 1  
    condition= cn % 2  
    
    if condition== 1:   
        return True
    else:
        return False

def get_inv_and_json(inventory, 
			filter_networks=[], 
			filter_stations=[],
			filter_domain=[-180,180,-90,90]):


	"""
	
	Uses fdsn to find available stations in a specific geographical location and time period.  
	Parameters
	----------
	json_path: str
		Path of the json file that will be returned
	providers:list
		list of Client object
	restrictions: DownloadRestrictions or MassDownloaderRestrictions
		Restrictions to download mseed
	from_xml : str
		Path of xml file
	**kwargs: str
		get_stations kwargs


	Returns
	----------
	stations_list.json: A dictionary containing information for the available stations.	  
		
	"""  
	if not filter_domain:
		filter_domain=[-180,180,-90,90]
		
	polygon = [(filter_domain[0],filter_domain[2]),
				(filter_domain[0],filter_domain[3]),
				(filter_domain[1],filter_domain[3]),
				(filter_domain[1],filter_domain[2]),
				(filter_domain[0],filter_domain[2])
				]

	station_list = {}   
	logger = logging.getLogger('json') 
	toprint = []
	stations_outside_domain = []
	for ev in inventory:
		net = ev.code
		if net not in filter_networks:
			for st in ev:
				sta = st.code
				msg = f'{net}-{sta}'
				logger.debug(msg)

				if sta not in filter_stations:

					elv = st.elevation
					lat = st.latitude
					lon = st.longitude

					is_in_domain = inside_the_polygon((lon,lat),polygon)

					if not is_in_domain:
						inventory = inventory.remove(network=net,station=sta)
						stations_outside_domain.append(sta)
						continue

					new_chan = [ch.code for ch in st.channels]

					if len(new_chan) > 0 and (sta not in station_list):
						channels = list(set(new_chan))
						sample_rates_gen = lambda x: st.select(channel=x).channels[0].sample_rate
						sample_rates = list(map(sample_rates_gen,channels))
						station_list[str(sta)] ={"network": net,
												"channels": list(set(new_chan)),
												"coords": [lat, lon, elv],
												"sampling_rate": sample_rates
												}

					toprint.append(msg)
				else:
					inventory = inventory.remove(network=net,station=sta)
		else:
			inventory = inventory.remove(network=net)
			
	logger.info(str(toprint) + ' ok')

	return inventory,station_list,stations_outside_domain

if __name__ == "__main__":
	from obspy.clients.fdsn import Client as FDSN_Client
	from obspy.core.utcdatetime import UTCDateTime
	# from .restrictions import PreprocRestrictions

	# IRIS_client = FDSN_Client(base_url="IRIS", user='gaprietogo@unal.edu.co',password="DaCgmn3hNjg")
	# st = IRIS_client.get_waveforms(network="YU",station="FC04,FC01",location="",channel="*",starttime=UTCDateTime("2016-04-22T00:00:00.0"),endtime=UTCDateTime("2016-04-23T00:00:00.0"))
	# st = get_st_according2preference(st,["","00","20","10"],["HH","BH"])
	# prove = get_all_sdswaveforms(client=IRIS_client,network="YU",station="FC04,FC01",location="",channel="HH?",starttime=UTCDateTime("2016-04-22T00:00:00.0"),endtime=UTCDateTime("2016-04-23T00:00:00.0"))
	# print(prove)


	# client = FDSN_Client('http://sismo.sgc.gov.co:8080')
	# inv = client.get_stations(network="CM",
	# 					  station="BAR2",
	# 					  location="*",
	# 					  channel="*",
	# 					  starttime=UTCDateTime("2019-04-23T00:00:00.0"),
	# 					  endtime=UTCDateTime("2019-04-23T00:02:00.0"))
	# st = client.get_waveforms(network="CM",
	# 					  station="BAR2",
	# 					  location="*",
	# 					  channel="*",
	# 					  starttime=UTCDateTime("2019-04-23T00:00:00.0"),
	# 					  endtime=UTCDateTime("2019-04-23T00:02:00.0"))
	# ppc_restrictions = PreprocRestrictions(["CM.BAR2"],detrend={'type':'simple'})
	# preproc_stream(st,ppc_restrictions)
	######### inventory
	# json_path = "/home/ecastillo/repositories/AIpicker_modules/onejson.json"
	# client_baseurl = "http://sismo.sgc.gov.co:8080"

	# restrictions = DownloadRestrictions(network="CM",
	#					   station="BAR2",
	#					   location="*",
	#					   channel="*",
	#					   starttime=UTCDateTime("2019-04-23T00:22:34.5"),
	#					   endtime=UTCDateTime("2019-04-25T00:23:39.5"),
	#					   chunklength_in_sec=5000,
	#					   overlap_in_sec=None,
	#					   groupby='{network}.{station}.{channel}')
	# xml = "/home/ecastillo/repositories/AIpicker_modules/CM.xml"

	# makeStationList(json_path,client_baseurl,restrictions,from_xml=xml)

	######## get stations
	# json_path = "/home/ecastillo/repositories/AIpicker_modules/onejson.json"
	# client_baseurl = "IRIS"
	# restrictions = DownloadRestrictions(network="CI",
	#				   station="BAK,ARV",
	#				   location="*",
	#				   channel="BH*",
	#				   starttime=UTCDateTime("2020-09-01 00:00:00.00"),
	#				   endtime=UTCDateTime("2020-09-02 00:00:00.00"),
	#				   chunklength_in_sec=3600,
	#				   overlap_in_sec=None,
	#				   groupby='{network}.{station}.{channel}')
	# makeStationList(json_path=json_path,client_baseurl="IRIS",restrictions=restrictions)


# /**
#  * @author [Emmanuel Castillo]
#  * @email [excastillot@unal.edu.co]
#  * @create date 2021-12-22 15:13:49
#  * @modify date 2021-12-22 15:13:49
#  * @desc [description]
#  */

"""
Concurrent downloader based on obspy's Mass Downloader class. 

Author:
    Emmanuel Castillo (ecastillot@unal.edu.co), 2020
"""

import time
import numpy as np
import itertools
import concurrent.futures 
from obspy.core.utcdatetime import UTCDateTime
from obspy.core.inventory.inventory import Inventory
from obspy.clients.fdsn.mass_downloader import (Restrictions,MassDownloader)
from SeisMonitor.utils import printlog

def _run_subprocess(mdl,domain,restriction,
                    mseed_storage,stationxml_storage):
    """
    Parameters
    ----------
    domain: class:'obspy.mass_downloader.domain'
        The download domain.
    restrictions: class:'obspy.mass_downloader.restrictions.Restrictions'
        Non-spatial downloading restrictions.
    mseed_storage: str
        Where to store the waveform files. 
    stationxml_storage: str
        Where to store the StationXML files.

    Returns:
    Download the mseed files.
    """
    try:
        tic = time.time()
        mdl.download(domain=domain,
                restrictions=restriction, 
                download_chunk_size_in_mb=20, 
                threads_per_client=3, 
                mseed_storage=mseed_storage,
                stationxml_storage=stationxml_storage)
        time.sleep(np.random.randint(25, 30))
        toc = time.time()

        printlog("info","Downloader",
        f"Thread done:{restriction.network} -- {restriction.station}"+
        f"\ttime: {round(toc-tic,2)} s")

    except Exception:
        printlog("error","Downloader",
        f"Thread failure: {restriction.network} -- {restriction.station}")
        pass

def process(args):
    """
    Parameters:
    -----------
    args: dict
        Contains the massdownloader parameters to download
        by process. 
    """

    mdl,domain,restriction,\
    mseed_storage,stationxml_storage = args

    try:
        tic = time.time()
        mdl.download(domain=domain,
                restrictions=restriction, 
                download_chunk_size_in_mb=20, 
                threads_per_client=3, 
                mseed_storage=mseed_storage,
                stationxml_storage=stationxml_storage)
        toc = time.time()
        printlog("info","Downloader",
        f"Process done:{restriction.network} -- {restriction.station}"+
        f"\ttime: {round(toc-tic,2)} s")                 

    except Exception:
        printlog("error","Downloader",
        f"Process failure: {restriction.network} -- {restriction.station}")
        pass

class MseedDownloader(object):
    def __init__(self,providers):
        self.providers= providers

    """Concurrent bulk downloader based on obspy's Mass Downloader class.

    parameters
    ----------
    providers: list
        list of Client instances

    returns
    -------
        MseedDownloader object

    Warnings:
    ---------
    client instances must have available the get_stations method    
    """

    def _get_stations_info(self,bulk):
        """
        Parameters:
        -----------
        bulk: str, file or list of lists
            Information about the requested data. 
            See get_stations_bulk from obspy for details.

        returns: 
        --------
            stations_info: list
            list of tuples with stations information
            (network,station)
        """

        inv = Inventory()
        for provider in self.providers:
            one_inv = provider.get_stations_bulk(bulk,level="station")
            # one_inv = provider.get_stations_bulk(bulk,level="response")
            inv = inv.__add__(one_inv) 

        stations_info = [(net.code, sta.code) 
                        for net in inv
                        for sta in net ]
        return stations_info

    def _build_station_restrictions(self,restrictions,
                                    stations_info):
        """
        Parameters:
        -----------
        restrictions: class:'obspy.mass_downloader.restrictions.Restrictions'
            Download restrictions from obspy
        stations_info: list
            list of tuples with stations information
            (network,station)

        Returns:
        --------
        rest_list: list
            List of restrictions by each tuple given by stations_info
        """
        rest_dict = restrictions.__dict__
        new_rest_dict = {}
        for key,value in rest_dict.items():
            if key[0] == '_':
                pass
            elif key == 'chunklength':
                new_rest_dict[key+'_in_sec'] = value
            else:
                new_rest_dict[key] = value

        rest_list = []
        for info in stations_info:
            net,sta = info

            new_rest_dict['network'] = net
            new_rest_dict['station'] = sta
            station_rest = Restrictions(**new_rest_dict)
            rest_list.append(station_rest)

        return rest_list

    def _prepare_args_for_process(self,domain,restrictions_list,
                                    mseed_storage,stationxml_storage):
        
        """
        method to join all massdownloader parameters and save it in 
        the list to map parallelly.
        """
        mdl = MassDownloader(providers=self.providers)
        args = zip(itertools.repeat(mdl), 
            itertools.repeat(domain),
            restrictions_list,
            itertools.repeat(mseed_storage),  
            itertools.repeat(stationxml_storage))
        return list(args)

    def download(self,domain,restrictions,
                    mseed_storage,stationxml_storage,
                    workers=None,parallel_mode="thread"):
        """
        Parallel download by each station. 

        Parameters
        ----------
        domain: class:'obspy.mass_downloader.domain'
            The download domain.
        restrictions: class:'obspy.mass_downloader.restrictions.Restrictions'
            Non-spatial downloading restrictions.
        mseed_storage: str
            Where to store the waveform files. 
        stationxml_storage: str
            Where to store the StationXML files.
        workers: int
            Number of subprocess that will be used.
        parallel_mode: str
            It can be 'thread' or 'process'. However It's recommended
            to use 'process' because by 'thread' you only can use 6 workers
            because for greater workers the downloading is not complete 
            due to no FDSN response.

        returns
        -------
        mseed files and xml files in mseed_storage and stationxml_storage
        respectively.
        """

        ## get the stations information and prepare the restrisctions_list
        ## for the parallel downloading

        bulk = [(restrictions.network, restrictions.station,
                '*', '*',
                 restrictions.starttime, restrictions.endtime) ]
        # bulk = [(restrictions.network, restrictions.station,
        #         restrictions.location, restrictions.channel,
        #          restrictions.starttime, restrictions.endtime) ]

        stations_info = self._get_stations_info(bulk)
        restrictions_list = self._build_station_restrictions(restrictions,
                                                            stations_info)


        mdl = MassDownloader(providers=self.providers)
        ## Go to the downloading: 
        if workers == 1:
            for rest in restrictions_list:
                args = (mdl,domain,rest,mseed_storage,stationxml_storage)
                process(args)
        elif workers == 0:
            raise Exception("workers must be grater than 1")
        else:
            total_tic = time.time()

            #Thread mode
            if parallel_mode in ("thread","t","T"):

                def subprocess(restriction):
                    _run_subprocess(mdl,domain,restriction,
                                    mseed_storage, stationxml_storage)

                with concurrent.futures.ThreadPoolExecutor(
                    max_workers=workers) as executor:
                    executor.map(subprocess,restrictions_list)

            #Process mode
            elif parallel_mode in ("process","p","P"):
                args = self._prepare_args_for_process(domain,
                                    restrictions_list,
                                    mseed_storage,
                                    stationxml_storage)
                with concurrent.futures.ProcessPoolExecutor(
                    max_workers=workers) as executor:
                    executor.map(process,args)
                
            else:
                raise Exception(f"Doesn't exist {parallel_mode} mode"
                                "only 1)thread or 2)process")

            total_toc = time.time()
            printlog("info","Downloader"," Total download time: "+
                    f"\t{round(total_toc-total_tic,2)} s")


if __name__ == "__main__":
    from obspy.clients.fdsn.mass_downloader.domain import RectangularDomain
    from obspy.clients.fdsn.client import Client


    IRIS_client = Client(base_url="IRIS", user="gaprietogo@unal.edu.co",
                        password="DaCgmn3hNjg")  

    CM_client = Client(base_url='http://10.100.100.232:8091')

    YU_restrictions = Restrictions(starttime=UTCDateTime(2016, 8, 1),
                            endtime=UTCDateTime(2016, 8, 2),
                            network="YU", 
                            station="CS01,CS02",
                            # location="*", channel="*",
                            chunklength_in_sec=86400,
                            reject_channels_with_gaps=False,
                            minimum_length=0.0,
                            minimum_interstation_distance_in_m=0.0,
                            channel_priorities=["HH[ZNE]", "BH[ZNE]","EH[ZNE]","HN[ZNE]"],
                            location_priorities=["", "00", "20", "10"])

    VMM_restrictions = Restrictions(starttime=UTCDateTime(2016, 8, 1),
                            endtime=UTCDateTime(2016, 8, 2),
                            network="CM", 
                            station="BAR2,VMM*",
                            # location="*", channel="*",
                            chunklength_in_sec=86400,
                            sanitize = False,
                            reject_channels_with_gaps=False,
                            minimum_length=0.0,
                            minimum_interstation_distance_in_m=1000,
                            channel_priorities=["HH[ZNE]", "BH[ZNE]","EH[ZNE]","HN[ZNE]"],
                            location_priorities=["", "00", "20", "10"])

    Colombian_domain = RectangularDomain(minlatitude=-2, maxlatitude=16,
                           minlongitude=-84, maxlongitude=-68)

    mseed_storage = "/home/ecastillo/download/prove/waveforms"
    stationxml_storage = "/home/ecastillo/download/prove/stations"

    mseed_dl = MseedDownloader([CM_client])
    mseed_dl.download(domain=Colombian_domain,
                        restrictions=VMM_restrictions, 
                        mseed_storage=mseed_storage,
                        stationxml_storage=stationxml_storage,
                        workers=1,parallel_mode="process",picker='phasenet')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Nov 12 20:00:00 2020
@author: Emmanuel_Castillo
last update: 14-11-2020 
"""


from obspy import UTCDateTime
from obspy.clients.fdsn import Client as FDSNClient

sgc_client = FDSNClient('http://10.100.100.13:8091')
st = sgc_client.get_waveforms(network="CM",
                    station="BAR2",
                    location="*",
                    channel="*",
                    starttime=UTCDateTime("2016-01-01T02:00:00.000000Z"),
                    endtime=UTCDateTime("2016-01-01T00:00:00.000000Z"))
print(st)


# /**
#  * @author [Emmanuel Castillo]
#  * @email [ecastillot@unal.edu.co]
#  * @create date 2022-03-17 08:53:23
#  * @modify date 2022-03-17 08:53:23
#  * @desc [description]
#  */

"""
This script is an example to make a client class
for specific data structure archive on local filesystem. 

The structure is:
"{root_path}","{field_name}","seedfiles",
"{year}-{month:02d}", "{year}-{month:02d}-{day:02d}", 
"{network}.{station}.{location}.{channel}.{year}.{julianday:03d}"

To do this, I builded OwnClient class as SDS client subclass. 
It  it inherits the SDS Client functionalities.
However field_name, month and day  are not used in fmt on the SDS client instance.
Therefore, it's mandatory override two private functions: 
_get_filenames, _get_filename.

The mandatory parameters for OwnClient class is: root_path and field_name
Example:
---------
root_path = "/home/emmanuel/myarchive"
client = OwnClient(root_path,"FIELD_1")
st = client.get_waveforms("YY","XXXX","00",
                        channel="HHZ",starttime = UTCDateTime("20220102T000100"),
                        endtime = UTCDateTime("20220102T000200"))
"""

import os 
import glob
from datetime import timedelta
from obspy.clients.filesystem.sds import Client
from obspy.core.util.misc import BAND_CODE

class LocalClient(Client):

    def __init__(self,root,fmt,**kwargs):
        """
        Parameters:
        -----------
        root: str
            Path where is located the Local structure
        fmt: str
            The parameter should name the corresponding keys of the stats object, e.g. 
            "{year}-{month:02d}/{year}-{month:02d}-{day:02d}/{network}.{station}.{location}.{channel}.{year}.{julday:03d}"

        **kwargs SDS client additional args
        """
        self.root = root
        self.fmt = fmt
        super().__init__(root,**kwargs)

    def _get_filenames(self, network, station, location, channel, starttime,
                       endtime, sds_type=None):
        """
        Get list of filenames for certain waveform and time span.
        :type network: str
        :param network: Network code of requested data (e.g. "IU").
        :type station: str
        :param station: Station code of requested data (e.g. "ANMO").
        :type location: str
        :param location: Location code of requested data (e.g. "").
        :type channel: str
        :param channel: Channel code of requested data (e.g. "HHZ").
        :type time: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param time: Time of interest.
        :type sds_type: str
        :param sds_type: None
        :rtype: str
        """
        sds_type = sds_type or self.sds_type
        # SDS has data sometimes in adjacent days, so also try to read the
        # requested data from those files. Usually this is only a few seconds
        # of data after midnight, but for now we play safe here to catch all
        # requested data (and with MiniSEED - the usual SDS file format - we
        # can use starttime/endtime kwargs anyway to read only desired parts).
        year_doy = set()
        # determine how far before starttime/after endtime we should check
        # other dayfiles for the data
        t_buffer = self.fileborder_samples / BAND_CODE.get(channel[:1], 20.0)
        t_buffer = max(t_buffer, self.fileborder_seconds)
        t = starttime - t_buffer
        t_max = endtime + t_buffer
        # make a list of year/doy combinations that covers the whole requested
        # time window (plus day before and day after)
        while t < t_max:
            year_doy.add((t.year,t.month,t.day, t.julday))
            t += timedelta(days=1)
        year_doy.add((t_max.year,t_max.month,t_max.day, t_max.julday))

        full_paths = set()
        for year,month,day,doy in year_doy:
            filename = self.fmt.format(
                            network=network, station=station, location=location,
                            channel=channel, year=year, month=month, 
                            day=day, julday=doy,sds_type=sds_type)
            full_path = os.path.join(self.sds_root, filename)
            full_paths = full_paths.union(glob.glob(full_path))
        
        return full_paths

    def _get_filename(self, network, station, location, channel, time, sds_type=None):
        """
        Get filename for certain waveform.
        :type network: str
        :param network: Network code of requested data (e.g. "IU").
        :type station: str
        :param station: Station code of requested data (e.g. "ANMO").
        :type location: str
        :param location: Location code of requested data (e.g. "").
        :type channel: str
        :param channel: Channel code of requested data (e.g. "HHZ").
        :type time: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param time: Time of interest.
        """
        sds_type = sds_type or self.sds_type
        filename = self.fmt.format(
                    network=network, station=station, location=location,
                    channel=channel, year=time.year, month=time.month, 
                    day=time.day, doy=time.julday,sds_type=sds_type)
        return os.path.join(self.sds_root, filename)

if __name__ == "__main__":
    from obspy import UTCDateTime

    # ### RSNC
    # root = "/home/emmanuel/RSNC-2016"
    # fmt = '{year}/{station}/{channel}.{sds_type}/{network}.{station}.{location}.{channel}.{sds_type}.{year}.{doy:03d}'
    # client = LocalClient(root,fmt)
    # st = client.get_waveforms("CM","BAR2","*",
    #                     channel="*Z",starttime = UTCDateTime("20160627T235400"),
    #                     endtime = UTCDateTime("20160627T235600"))

    archive = "/home/emmanuel/Descargas/SeisMonitor_dataset/archive/mseed"
    my_local_fmt = os.path.join("{year}-{month:02d}", 
                    "{year}-{month:02d}-{day:02d}", 
                    "{network}.{station}.{location}.{channel}.{year}.{julday:03d}")
    carma_client = LocalClient(archive,my_local_fmt)
    st = carma_client.get_waveforms(network="YU",
                        station="GJ*",
                        location="*",
                        channel="*",
                        starttime=UTCDateTime("2017-12-24T00:00:00.000000Z"),
                        endtime=UTCDateTime("2017-12-24T00:10:00.000000Z"))
    print(st)






import sys
import os
seismopath = "/home/emmanuel/EDCT"
seismonitor = os.path.join(seismopath,"SeisMonitor")
sys.path.insert(0,seismonitor)

from datetime import datetime
import numpy as np
import pandas as pd
from tqdm import tqdm
from glob import glob
from multiprocessing import Pool
from obspy import read_inventory
import time
from obspy.core.event.catalog import Catalog, read_events
import pexpect
import subprocess
from SeisMonitor.utils import printlog, isfile
# import SeisMonitor.utils as ut

# if 'google.colab' in sys.modules:
#     __file__ =

libgfortran_path = os.path.join(os.path.dirname(__file__),"libgfortran.so.3.0.0")
CORE_SEISAN = os.path.join(os.path.dirname(__file__),"core")
SEISAN_path = os.path.join(CORE_SEISAN,"seismo")
COM_path = os.path.join(SEISAN_path,"COM")
PRO_path = os.path.join(SEISAN_path,"PRO")
DAT_path = os.path.join(SEISAN_path,"DAT")
STATION0_path = os.path.join(DAT_path,"STATION0.HYP")


def download_seisan(libgfortran_path):
    '''
    Seisan can be downloaded from https://www.geo.uib.no/seismo/SOFTWARE/SEISAN/seisan_v12.0_linux_64.tar.gz
    '''
    name = "seisan_v12.0_linux_64.tar.gz"
    gz_path = os.path.join(CORE_SEISAN,name)

    if not os.path.isdir(CORE_SEISAN):
      os.makedirs(CORE_SEISAN)
    if not os.path.isdir(gz_path):
      os.system(f"wget -O {gz_path} https://www.geo.uib.no/seismo/SOFTWARE/SEISAN/seisan_v12.0_linux_64.tar.gz")
    
    os.system("sudo apt-get update")
    os.system("sudo apt-get install gcc")
    os.system("sudo apt-get install g++")
    os.system("sudo apt-get install gfortran")

    if not os.path.isdir(SEISAN_path):
        os.makedirs(SEISAN_path)
        os.system(f"tar -xf {gz_path} -C {SEISAN_path}")
    
    os.system(f"cp {libgfortran_path} /usr/lib/x86_64-linux-gnu/libgfortran.so.3")

def split(sfile_folder,sfilename):
    """
    Parameters:
    -----------
    sfile_folder: str
        It is the folder where is located the sfile.
        In addition, it is the folder where will be saved all sfiles.

    sfilename: str
        Name of the sfile what will be splitted

    Returns:
    --------
        Each sfile produced by the split function

    """
    split_path = os.path.join(PRO_path,"split")
    Split = pexpect.spawn(split_path, cwd=sfile_folder)
    # Split = pexpect.spawn('split', cwd=sfile_folder)
    Split.expect(b' NAME')
    Split.sendline(f'{sfilename}\n')
    Split.expect(b'LOCAL DIRECTORY')
    Split.sendline('\n')
    Split.expect(b'CHARS')
    Split.sendline('usr\n')
    # print(Split.before)
    Split.interact()

def collect(sfile_folder):
    """
    Parameters:
    -----------
    sfile_folder: str
        It is the folder where is located all sfiles.
        In addition, it is the folder where will be saved 
        the collected sfile named as collect.out
    
    Returns:
        colelcted sfile in the next path
        {sfile_folder}/collect.out

    
    """
    collect_path = os.path.join(PRO_path,"collect")
    Collect = pexpect.spawn(collect_path, cwd=sfile_folder)
    Collect.expect(b' return for default')
    Collect.sendline(',,\n')
    Collect.expect(b":")
    Collect.sendline('\n')
    Collect.expect(b":")
    Collect.sendline('\n')
    Collect.expect(b"(Y/N=default)")
    # print(Collect.before)
    Collect.interact()

def update(sfile_folder):
    """
    Parameters:
    -----------
    sfile_folder: str
        It is the folder where is located all sfiles.
            
    Returns:
        It updates each sfile with the hypocenter location
    """
    update_path = os.path.join(PRO_path,"update")
    Update = pexpect.spawn(update_path, cwd=sfile_folder)
    Update.expect(b' return for default')
    Update.sendline(',,\n')
    Update.expect(b":")
    Update.sendline('usr\n')
    Update.expect(b":")
    Update.sendline('\n')
    # print(Update.before)
    Update.interact()

def select(sfile_folder):
    """
    Parameters:
    -----------
    sfile_folder: str
        It is the folder where is located all sfiles.
            
    Returns: 
        It executes select command from SEISAN.
        select.in: index
        select.out: Selected information
    """
    select_path = os.path.join(PRO_path,"select")
    Select = pexpect.spawn(select_path, cwd=sfile_folder)
    Select.expect('FILENAME FOR ONE FILE, MUST BE 6 OR MORE CHARACTERS OR HAVE A .')
    Select.sendline(',,\n')
    Select.expect(' yyyymmddhhmmss:')
    Select.sendline('\n')
    Select.expect(' :')
    Select.sendline('\n')
    Select.expect('RETURN TO SEARCH:')
    Select.sendline('\n')
    # print(Select.before)
    Select.interact()

    out = os.path.join(sfile_folder,"select.out ")
    os.system(f'head -1 {out}')

def norhead(sfile_folder):
    """
    Parameters:
    -----------
    sfile_folder: str
        It is the folder where is located all sfiles.
            
    Returns:
        It select the principal header information of the event in
        only one file
    """
    norhead_path = os.path.join(PRO_path,"norhead")
    select_out = os.path.join(sfile_folder,"select.out")
    hypo_out = os.path.join(sfile_folder,"hypo.out")
    msg = f"{norhead_path} {select_out} {hypo_out}"
    # print(msg)
    os.system(msg)

def cp_station0(station0,sfile_folder):
    """
    Parameters:
    -----------
    station0: str
        Path of the station0 file
    sfile_folder: str
        It is the folder where is located all sfiles.
            
    Returns:
        Copy the station0 file in the sfile_folder
    """
    station0_path = os.path.join(DAT_path,"STATION0.HYP")
    if os.path.isfile(station0) != True:

        msg = f"cp {station0} {sfile_folder}"
        # print(msg)
        os.system(msg)

def resp2df(resp):
    """
    Parameters:
    -----------
    resp: str
        RESP filepath

    Returns: DataFrame
        Dataframe with the next columns
        network,station,latitude,longitude,elevation
    """
    networks = []
    stations = []
    longitudes = []
    latitudes = []
    elevations = []
    inv = read_inventory(resp)
    for net in inv:
        for sta in net:
            latitudes.append(sta.latitude)
            longitudes.append(sta.longitude)
            elevations.append(sta.elevation)
            stations.append(sta.code)
            networks.append(net.code)

    df = {"network":networks,"station":stations,
        "latitude":latitudes,"longitude":longitudes,
        "elevation":elevations}
    df = pd.DataFrame(df)
    # print(df)
    return df

def sta2station0(df):
    """
    Parameters:
    ----------
    df: DataFrame
        Dataframe with the next columns
        network,station,latitude,longitude,elevation.
        Review resp2df function.

    Returns: 
    --------
    msgs: list
        List of each message by station.
        The station message contains the next information:
        
        station lat lon elevation
        ---example---
        CA01A 353.16N 7341.07W 436

    """
    df = df.sort_values(by="station")

    df["lon_dec"] = df["longitude"].apply(lambda x: round((abs(x) % 1)*60,2))
    df["lat_dec"] = df["latitude"].apply(lambda x: round((abs(x) % 1)*60,2))
    df["latitude"] = df["latitude"].apply(lambda x: round(int(x),2))
    df["longitude"] = df["longitude"].apply(lambda x: round(int(x),2))

    msgs = []
    for i,row in df.iterrows():
        sta = row.station
        lat = str(abs(row.latitude)).zfill(2)
        lon = str(abs(row.longitude)).zfill(2)

        if row.latitude < 0:
            latcord = "S"
        else:
            latcord = "N"

        if row.longitude < 0:
            loncord = "W"
        else:
            loncord = "E"

        fmt = "%6s%2i%5.2f%1s%3i%5.2f%1s%4i\n"
        msg = fmt % (sta,int(lat),float(row.lat_dec),
                    latcord,int(lon),float(row.lon_dec),
                    loncord,int(row.elevation))
        msgs.append(msg)
        # print(msg)
    return msgs

def vel2station0(df,only_vp=True):
    """
    Parameters:
    -----------
    df: DataFrame
        Dataframe with the next columns
        dep,vp,vs,disc

    Returns:
    --------
    msgs: list
        List of each message by discontinuity in the velocity model.
        The discontinuity message contains the next information:
        
        vp dep vs disc
        ---example---
        7.000 25.000  3.940B

    Warnings:
    ---------
        Discontinuities
        B: Conrad
        N: Moho
    """
    df = df.sort_values(by="vp")
    df = df.where(pd.notnull(df), None)
    msgs = []
    for i,row in df.iterrows():
        if row.disc != None:
            if only_vp:
                fmt = "%7.3f%7.3f%1s\n"
                msg = fmt % (row.vp,row.depth,row.disc)
            else:
                fmt = "%7.3f%7.3f%7.3f%1s\n"
                msg = fmt % (row.vp,row.depth,row.vs,row.disc)
        else: 
            if only_vp:
                fmt = "%7.3f%7.3f\n"
                msg = fmt % (row.vp,row.depth)
            else:
                fmt = "%7.3f%7.3f%7.3f\n"
                msg = fmt % (row.vp,row.depth,row.vs)

        # print(msg)
        msgs.append(msg)
    return msgs

def test2station0(test):
    """
    Parameters:
    -----------
    test: dict
        key: str
            Number of the SEISAN HYPOCENTER test line
        value: int
            Value of the respective key

    Returns:
    --------
    msgs: list
        List of each message by test line.

        The test message contains the next information:
        RESET TEST(key)=value

        ------example-------
        RESET TEST(02)=500.0
        
    """
    msgs = []
    for key,value in test.items():
        key = str(key).zfill(2)
        msg = f"RESET TEST({key})={value}\n"
        # print(msg)
        msgs.append(msg)

    return msgs

def vsp_ratio2station0(vsp_ratio):
    """
    Parameters:
    -----------
    vsp_ratio: ordered dict
        The key order of the dictionary must be the next:
        "starting_depth","xnear","xfar","vps"

        ---------example--------
        {"starting_depth":2,
        "xnear":100,
        "xfar": 800,
        "vps": 1.84},

    Returns:
    --------
    msg: str
        Represents the last line of the STATION0 file

    """
    fmts = []
    values = []
    for key,value in vsp_ratio.items():
        if key == "vps":
            fmt = "5.2f"   
        else:
            fmt = "5.0f"
            
        fmts.append(fmt)
        values.append(float(value))

    fmt = "%"+"%".join(fmts)
    # print(fmt)
    # print(tuple(values))
    msg = fmt % tuple(values) + "\n"
    # print(msg)
    return msg

def check_sfile_integrity(sfile_folder,rm_not_locatable=True):
    """
    It checks sfile integrity.
    We aim that obspy could be read it as catalog.

    Parameters:
    -----------
    sfile_folder: str
        It is the folderpath where is located all sfiles.
    rm_not_locatable: bool
        Removes not locatable events
    """
    for dp, dn, filenames in os.walk(sfile_folder):
        for f in filenames:
            # print(f)
            #check if is an sfile to collect
            if f.split(".")[-1][0] == "S":

                sfilepath = os.path.join(dp, f)

                with open(sfilepath,"r") as sf:
                    line0 = sf.readline()
                    # print(line0[0:10])
                    
                    latlondep_col = line0[23:45]

                    if latlondep_col.isspace():
                        print(sfilepath,"not_locatable")
                    
                        if rm_not_locatable:
                            os.remove(sfilepath)
                            print("Removed:",sfilepath,"not_locatable")

class STATION0():
    def __init__(self,xml_path, vel_path,
            test={"02":500.0,"11":99.0,
                "13":1.0,"41":20000.0,
                "43":5.0,"56":1.0,
                "85":0.1},
            vsp_ratio = {"starting_depth":2,
                        "xnear":100,
                        "xfar": 800,
                        "vps": 1.78},
            agency = "TES",
            only_vp = True
                ):
        """
        Parameters:
        -----------
        sta_df: DataFrame
            Dataframe with the next columns
            network,station,latitude,longitude,elevation.
            Review resp2df function.
        vel_df: DataFrame
            Dataframe with the next columns
            dep,vp,vs,disc
        test: dict
            key: str
                Number of the SEISAN HYPOCENTER test line
            value: int
                Value of the respective key
        vsp_ratio: ordered dict
            The key order of the dictionary must be the next:
            "starting_depth","xnear","xfar","vps"

            ---------example--------
            {"starting_depth":2,
            "xnear":100,
            "xfar": 800,
            "vps": 1.84},
        agency: str
            Agency 
        """

        self.test = test
        self.sta_df = resp2df(xml_path)
        self.vel_df = pd.read_csv(vel_path)
        self.vsp_ratio = vsp_ratio
        self.agency = agency
        self.only_vp = only_vp
    
    def _get_msgs(self):
        """
        Get the messages to write in the station0 file
        """
        test_msgs = test2station0(self.test) 
        sta_msgs = sta2station0(self.sta_df)
        vel_msgs = vel2station0(self.vel_df,only_vp=self.only_vp)
        vsp_msg = vsp_ratio2station0(self.vsp_ratio)


        msgs = test_msgs + ["\n"] +sta_msgs +\
                ["\n"] + vel_msgs + ["\n"] +\
                [vsp_msg] + [self.agency]
        # print(msgs)

        return msgs

    def write(self,out):
        """
        Parameters:
        -----------
        out: str
            Station0 path file
        """
        # ut.isfile(out)
        f = open(out, "w")
        for line in self._get_msgs():
            f.write(line)
        f.close()

class HypocenterTools():
    def __init__(self,sfile_folder,from_sfilename=None):
        """
        Parameters:
        -----------
        sfile_folder: str
            It is the folderpath where is located all sfiles.
        station0: str
            Station0 path
        from_sfilename: str
            Name of the only one sfile with specific filename
            that you want to locate.
        """
        self.sfile_folder = sfile_folder
        self.sfilename = from_sfilename

    def remodl_and_setbrn(self):
        """
        Copy station0 in sfiles folder
        """
        subprocess.Popen(["remodl","setbrn"],cwd=self.sfile_folder)

        # cp_station0(self.station0_base,self.sfile_folder)
    
    def update(self):
        """
        Locate the events with the update seisan command
        """
        update(self.sfile_folder)
    
    def collect(self):
        """
        Collect all sfiles in only one collect sfile and write it
        in {out} path with the {format} specified.

        Parameters:
        -----------
        out: str
            Output Path of the collected catalog
        format: str
            Format supported by OBspy catalogs
        """
        collect(self.sfile_folder)
        nordic = os.path.join(self.sfile_folder,"collect.out")
        ## check if nordic event is ok
        catalog = read_events(nordic,format="NORDIC")
        ## put network and channel in catalog

        # # ut.isfile(out)
        # catalog.write(out,format=format)
        return catalog

    def select(self):
        """
        Runs select seisan command
        """
        select(self.sfile_folder)

    def norhead(self):
        """
        Runs norhead seisan command
        """
        norhead(self.sfile_folder)

    def split(self):
        """
        Runs split seisan command
        """
        split(self.sfile_folder,self.sfilename)



if __name__ == "__main__":
    # download_seisan()

    # xml_path = "/home/emmanuel/EDCT/SeisMonitor/data/metadata/CM.xml"
    # vel_path = "/home/emmanuel/EDCT/SeisMonitor/data/metadata/vel1d_col.csv"

    # sta0 = STATION0(xml_path,vel_path)
    # sta0.write("./STATION0.HYP")

    # seisan_folder= "/home/emmanuel/EDCT/SeisMonitor/SeisMonitor/monitor/locator/hypocenter/test/"
    # sfilename = "12-2223-00L.S202106"
    # split(seisan_folder,sfilename )
    # update(seisan_folder )


    ####### cse
    xml_path = "/home/emmanuel/G-Ecopetrol/ecastillo/Avances/2022/cano_sur_datos_iniciales/CSE.xml"
    vel_path = "/home/emmanuel/G-Ecopetrol/ecastillo/Avances/2022/cano_sur_datos_iniciales/vel_model.csv"

    sta0 = STATION0(xml_path,vel_path)
    sta0.write("./STATION0.HYP")


# /**
#  * @author [Emmanuel Castillo]
#  * @email [excastillot@unal.edu.co]
#  * @create date 2021-12-22 09:49:23
#  * @modify date 2021-12-22 09:49:23
#  * @desc [description]
#  */

SeisMonitor_columns = ["pick_id","arrival_time","probability","phasehint",
                    "network","station","location","instrument_type","author",
                    "creation_time","event_start_time","event_end_time",
                    "detection_probability","snr","station_lat","station_lon",
                    "station_elv","file_name"]

                    #phasenet other cols
                    # "mseed_start_time","mseed_end_time","sampling_rate","sample",
                    # "segment","segment_type"]

"""
Utils file to use EQTransformer Monitor
"""


import os
import csv
import sys
import json
import math
import time
import obspy
import shutil
import logging
import datetime
import numpy as np
import pandas as pd
import datetime as dt
from obspy import UTCDateTime
from datetime import timedelta
from SeisMonitor.utils import printlog,isfile
from SeisMonitor.monitor.downloader.utils import get_chunktimes
import concurrent.futures as cf
from obspy.core.event.origin import Pick
from obspy.core.event.base import (QuantityError,
                                WaveformStreamID,
                                CreationInfo,
                                Comment)
from git import Repo
from obspy.core.event import ResourceIdentifier


def git_clone_aipicker(name,repo_dir):
    """
    Params:
    -------
    name: str
        EQTransformer or PhaseNet
    repo_dir: str
        Directory path to place the repository
    """
    if name == "PhaseNet":
        git_url = "https://github.com/ecastillot/PhaseNet.git"
    elif name == "EQTransformer":
        git_url = "https://github.com/ecastillot/EQTransformer.git"
    else:
        return False

    repo_dir = os.path.join(repo_dir,name)
    if os.path.isdir(repo_dir):
        print("There is alaready")
        return True
    else:
        Repo.clone_from(git_url, repo_dir)
        return True

## EQTransformer functions
def eqt_picks_2_seismonitor_fmt(eqt_folder,mseed_folder,out_path):
    #eqt
    date_cols = ["p_arrival_time","s_arrival_time",
                "event_start_time","event_end_time"]

    dfs = []
    for dp, dn, filenames in os.walk(eqt_folder):
        for f in filenames:
            if f == "X_prediction_results.csv" :
                search_path = os.path.join(dp, f)
                df = pd.read_csv(search_path)
                if not df.empty:
                    df[date_cols] = df[date_cols].apply(pd.to_datetime)
                    dfs.append(df)

    if not dfs:
        return pd.DataFrame()

    df = pd.concat(dfs,ignore_index=True)
    df["station"] = df["station"].apply(lambda x: x.strip())
    df = df.sort_values(by="p_arrival_time",ignore_index=True)

    x_results_path = os.path.join(os.path.dirname(out_path),
                                "X_prediction_results_merge.csv")
    df.to_csv(x_results_path,index=False,
                date_format="%Y-%m-%d %H:%M:%S.%f")

    #seismonitor
    date_cols = ["arrival_time","creation_time",
                "event_start_time","event_end_time"]

    get_loc = lambda x: x.file_name.split(".")[2]
    df["location"] = df.apply(get_loc,axis=1)
    df["author"] = "EQTransformer"

    filename_func = lambda x: os.path.join(mseed_folder,x)
    df["file_name"] = df["file_name"].apply(filename_func)

    p_cols = df.columns.to_list()
    s_cols = df.columns.to_list()
    dfs = []
    for phase in ["p","s"]:
        if phase.lower() == "p":
            rm = "s"
        else:
            rm = "p"

        phase_cols = df.columns.to_list()
        not_rm_cols = [f'{phase}_arrival_time', f'{phase}_probability',
                         f'{phase}_uncertainty', f'{phase}_snr']
        rm_cols = [f'{rm}_arrival_time', f'{rm}_probability',
                 f'{rm}_uncertainty', f'{rm}_snr']
        for x in rm_cols:
            phase_cols.remove(x)

        phase_df = df[phase_cols]

        new_cols_gen = lambda x: x[2:]
        new_cols = list(map(new_cols_gen,not_rm_cols))

        columns = dict(zip(not_rm_cols, new_cols))
        phase_df = phase_df.rename(columns=columns)
        phase_df["phasehint"] = phase.upper()
        phase_df = phase_df.dropna(subset=["arrival_time"])

        phase_df["creation_time"] = dt.datetime.now()
        pick_id = lambda x: id_maker(x.arrival_time, x.network, 
                            x.station, x.location, 
                            x.instrument_type, x.phasehint)
        phase_df["pick_id"] = phase_df.apply(pick_id,axis=1) 


        phase_df[date_cols] = phase_df[date_cols].apply(pd.to_datetime)

        dfs.append(phase_df)
        

    df = pd.concat(dfs,ignore_index=True)
    df = df[SeisMonitor_columns]
    df = df.sort_values(by="arrival_time",ignore_index=True)
    df.to_csv(out_path,index=False,
            date_format="%Y-%m-%d %H:%M:%S.%f")
    return df
## PhaseNet functions
def get_filenames(mseed, filter_net=[],
                filter_sta=[],filter_cha=[]):
    """
    Parameters:
    -----------
    mseed: str
        path of the folder where you need to get all mseed filenames.
    filter_net: list
        avoid mseed if its filename contain one network specified in
        filter net
    filter_sta: list
        avoid mseed if its filename contain one station specified in
        filter net
    filter_cha: list
        avoid mseed if its filename contain one channel specified in
        filter net
    Returns:
    --------
    list
    Get all filenames contained in mseed folder.
    """

    if not isinstance(filter_net, list):
        raise Exception("filter_net is a list")
    elif not isinstance(filter_sta, list):
        raise Exception("filter_sta is a list")
    elif not isinstance(filter_cha, list):
        raise Exception("filter_cha is a list")

    filenames = []
    file_list = [ev for ev in os.listdir(mseed) \
                if ev.split('/')[-1].split('.')[-1].lower() == 'mseed']
    for onefile in file_list:
        filename = onefile.split('/')[-1]
        info = filename.split('__')[0]
        net,sta,loc,cha = info.split('.')
        if net in filter_net:
            pass
        if sta in filter_sta:
            pass
        if len(filter_cha) != 0:
            if cha not in filter_cha:
                pass
        else:
            filenames.append(filename)
    return filenames

def mv_mseed2onefolder(mseed_folder,one_folder):
    # print("aca",mseed_folder)
    # print("aca",one_folder)
    for dp, dn, filenames in os.walk(mseed_folder):
        for f in filenames:
            if f.endswith(".mseed") :
                mseed_file = os.path.join(dp, f)
                moved_mseed_file = os.path.join(one_folder, f)
                # os.copy()
                # print(mseed_file, moved_mseed_file)
                printlog("debug","mv_mseed2onefolder",f"{mseed_file} -> {moved_mseed_file}")
                shutil.move(mseed_file, moved_mseed_file)

def mv_mseed2stationfolder(one_folder,mseed_folder):
    for dp, dn, filenames in os.walk(one_folder):
        for f in filenames:
            if f.endswith(".mseed") :
                mseed_file = os.path.join(dp, f)

                src_id = f.split("__")[0]
                net,sta,loc,cha = src_id.split(".")

                moved_mseed_file = os.path.join(mseed_folder, sta,f)
                if not os.path.isdir(os.path.dirname(moved_mseed_file)):
                    os.makedirs(os.path.dirname(moved_mseed_file))
                # os.copy()
                printlog("debug","mv_mseed2stationfolder",f"{mseed_file} -> {moved_mseed_file}")
                # print(mseed_file, moved_mseed_file)
                shutil.move(mseed_file, moved_mseed_file)

def make_dataframe(mseed,json_path, filter_net=[],
                filter_sta=[],filter_cha=[]):
    with open(json_path, "r") as jfile:
        datajson = json.load(jfile)

    
    filenames = get_filenames(mseed, filter_net,
                filter_sta,filter_cha)
    logger = logging.getLogger('PhaseNet: datalist')
    
    mseed_paths = []
    for onefile in filenames:
        df = {}
        path = os.path.join(mseed,onefile)
        st = obspy.read(path)
        stats = st[0].stats

        for key,val in st._get_common_channels_info().items():
            net,sta,loc,instrument_type = key
            channels = list(val["channels"].keys())

        df["fname"] = path

        for channel in channels:
            direction = channel[-1]
            if len(channels) == 1:
                df["E"] = channel
                df["N"] = channel
                df["Z"] = channel
            else:
                df[direction] = channel
        

        instrument = instrument_type[:-1]
        instruments = list(map(lambda x: x[:-1], datajson[sta]['channels']))
        my_samplingrate = datajson[sta]['sampling_rate']
        idxs = [i for i,x in enumerate(instruments) if x == instrument]
        if not idxs:
            continue
        else:
            sampling_rate = my_samplingrate[idxs[0]]
        lat,lon,elv = datajson[sta]['coords']

        df['mseed_start_time'] = stats.starttime.strftime("%Y-%m-%d %H:%M:%S.%f")
        df['mseed_end_time'] = stats.endtime.strftime("%Y-%m-%d %H:%M:%S.%f")
        df["network"] = net
        df["station"] = sta
        df["location"] = loc
        df["instrument_type"] = instrument
        df["sampling_rate"] = sampling_rate
        df["sta_lat"] = lat
        df["sta_lon"] = lon
        df["sta_elv"] = elv

        mseed_paths.append(df)
    df = pd.DataFrame(mseed_paths)
    return df

def make_PhaseNet_datalist(datadir, json_path, datalist_path, channel_list=[], 
                            filter_network=[], filter_station=[],**kwargs ):

    df = make_dataframe(datadir,json_path, filter_network,
                        filter_station,channel_list)
    datalist_dir = os.path.dirname(datalist_path)
    isfile(datalist_path)
    df.to_csv(datalist_path,index=False)

    return datalist_dir

def create_datalist(self,all_in_folder=False):
    tic = time.time()
    logger = logging.getLogger('PhaseNet: datalist')
    logger.info("Running to create datalist")
    
    if not os.path.exists(self.datalist_dir):
        os.makedirs(self.datalist_dir)

    if all_in_folder:
        datalist = os.path.join(self.datalist_dir,'fname.csv')
        datalist = make_PhaseNet_datalist(datadir=self.all_mseed, 
                                datalist_path=datalist, 
                                groupby='{network}.{station}')

    else:

        stations = [ sta for sta in os.listdir(self.mseed_storage)]

        for sta in stations:
            datadir = os.path.join(self.mseed_storage,sta)
            datalist_path = os.path.join(self.datalist_dir,sta,'fname.csv')
            datalist = make_PhaseNet_datalist(datadir=datadir, 
                                    datalist_path=datalist_path, 
                                    groupby='{network}.{station}.{location}')
    
    toc = time.time()
    exetime = timedelta(seconds=toc-tic)
    logger.info(f'Total time of execution: {exetime.total_seconds()} seconds')

def phasenet_from_console(pnet_obj,msg_author):
    phasenet_dir = os.path.dirname(os.path.dirname(pnet_obj.model_dir))
    run = os.path.join(phasenet_dir,"run.py")

    ### nt for windows
    if os.name == 'nt':
        command = f"python \"{run}\" --mode=pred \
            --model_dir=\"{pnet_obj.model_dir}\" --data_dir=\"{pnet_obj.data_dir}\" \
            --data_list=\"{pnet_obj.data_list}\" --output_dir=\"{pnet_obj.output_dir}\"\
            --batch_size={pnet_obj.batch_size} --tp_prob={pnet_obj.tp_prob}\
            --ts_prob={pnet_obj.ts_prob} --input_mseed"

    else:
        command = f"python {run} --mode=pred \
            --model_dir={pnet_obj.model_dir} --data_dir={pnet_obj.data_dir} \
            --data_list={pnet_obj.data_list} --output_dir={pnet_obj.output_dir}\
            --batch_size={pnet_obj.batch_size} --tp_prob={pnet_obj.tp_prob}\
            --ts_prob={pnet_obj.ts_prob} --resampling={pnet_obj.one_single_sampling_rate} --input_mseed"

    if pnet_obj.plot_figure == True:
        command += ' ' + '--plot_figure' +' '+'' 
    if pnet_obj.save_result == True:
        command += ' ' + '--save_result'
    # print(command)
    printlog("info",msg_author,"Console-> "+command)
    printlog("info",msg_author,"Running...")
    os.system(command)

def get_pickframe(path,csv_filename):
    """
    Parameters:
    -----------
    path: str
        Path of the directory that contain the stations folders with
        the picks.
    csv_filename: str
        Name of the file that contains the picks in each station folder
    """
    pick_paths = []
    for dp, dn, filenames in os.walk(path):
        for f in filenames:
            if f == csv_filename:
                pick_path = os.path.join(dp, f)
                pick_paths.append(pick_path)
    li = []
    for pick_path in pick_paths:
        df = pd.read_csv(pick_path, index_col=None, header=0)
        li.append(df)

    if len(li) == 1:
        pickframe = li[0]
    else:
        pickframe = pd.concat(li, axis=0, ignore_index=True)
    return pickframe

def merge_picks(path, csv_filename, output_path, sort=None):
    """
    Parameters:
    -----------
    path: str
        Path of the directory that contain the stations folders with
        the picks.
    csv_filename: str
        Name of the file that contains the picks in each station folder
    """
    pickframe = get_pickframe(path,csv_filename)
    if sort != None:
        pickframe = pickframe.sort_values(sort)
    
    if os.path.isdir(os.path.dirname(output_path)) == False:
        os.makedirs(os.path.dirname(output_path))

    pickframe.to_csv(output_path,index=False)

def rm_phasenet_duplicate_picks(path,output_path):
    
    def select_pick(df):
        logger = logging.getLogger('PhaseNet: Select_picks')

        def average(df):
            probs = df['probability'].to_numpy()
            return np.average(probs)

        df_overlap = df[ df['segment_type'] == 'overlap']
        df_single = df[ df['segment_type'] == 'single']

        if df_overlap.empty == True:
            logger.debug(f'Single selected')
            return df_single
        elif df_single.empty == True:
            logger.debug(f'Overlap selected')
            return df_overlap
        else:
            single_prob_avg = average(df_single)
            overlap_prob_avg = average(df_overlap)

            if single_prob_avg >= overlap_prob_avg:
                logger.debug(f'Single:{single_prob_avg} >= Overlap:{overlap_prob_avg} | Single selected')
                return df_single
            else:
                logger.debug(f'Overlap:{overlap_prob_avg} > Single:{single_prob_avg} | Overlap selected')
                return df_overlap
    
    df = pd.read_csv(path,dtype={"location":str})
    if df.empty == True:
        return df

    # df = df[ df['time'].between(starttime,endtime) ]
    cols = ['mseed_start_time','mseed_end_time','sampling_rate','segment','sample','segment_type']
    strftime = "%Y-%m-%d %H:%M:%S.%f"
    timecols = ['arrival_time','creation_time','mseed_start_time','mseed_end_time']
    df[timecols] = df[timecols].apply(pd.to_datetime)


    mygroup = ['mseed_start_time','mseed_end_time','network',\
                'station','instrument_type']
    # mygroup = ['mseed_start_time','mseed_end_time','network',\
    #             'station','instrument_type','author','snr',
    #             'station_lat','station_lon','station_elv']

    mseed_groups = df.groupby(mygroup,as_index=False)
    indexes = list(mseed_groups.indices.keys())

    picks_ok = []
    for ind in indexes:
        mseed_df = mseed_groups.get_group(ind)
        mseed_starttime = mseed_df['mseed_start_time'].iloc[0] 
        mseed_endtime = mseed_df['mseed_end_time'].iloc[0] 
        mseed_sr = mseed_df['sampling_rate'].iloc[0]

        # 1500 samples (phasenet) / sampling_rate of the mseed file
        overlap_in_sec =  1500 / mseed_sr 

        overlap_times = get_chunktimes(mseed_starttime,mseed_endtime,
                        overlap_in_sec,0)

        for ovl_time in overlap_times:
            ovl_starttime, ovl_endtime = ovl_time
            ovl_df = mseed_df[ (mseed_df['arrival_time'] <= ovl_endtime) & \
                               (mseed_df['arrival_time'] >= ovl_starttime)  ]
            if ovl_df.empty == True:
                pass
            else:
                pick_ok = select_pick(ovl_df)
                picks_ok.append(pick_ok)
                
    df_picks_ok = pd.concat(picks_ok,ignore_index=True,sort=False)
    df_picks_ok = df_picks_ok.sort_values('arrival_time')
    df_picks_ok = df_picks_ok.reset_index(drop=True)
    df_picks_ok.to_csv(output_path,index=False)
    
    return df_picks_ok

def picks2df(picks):
    df = []
    for pick in picks:
        seismonitor = {}
        
        seismonitor["pick_id"] = pick.resource_id.id
        seismonitor["arrival_time"] = pick.time.strftime("%Y-%m-%d %H:%M:%S.%f")
        seismonitor["probability"] = pick.time_errors.uncertainty
        seismonitor["phasehint"] = pick.phase_hint 
        seismonitor["network"] = pick.waveform_id.network_code
        seismonitor["station"] = pick.waveform_id.station_code
        seismonitor["instrument_type"] = pick.waveform_id.channel_code[0:2]
        seismonitor["author"] = pick.method_id.id
        seismonitor["creation_time"] = pick.creation_info.creation_time.strftime("%Y-%m-%d %H:%M:%S.%f")
        
        if isinstance(pick.waveform_id.location_code,int):
            if math.isnan(pick.waveform_id.location_code):
                seismonitor["location"] = ""
        elif pick.waveform_id.location_code == None:
            seismonitor["location"] = ""
        else:
            seismonitor["location"] = "{:02d}".format(int(pick.waveform_id.location_code))

        for key,val in pick.comments[0].items():
            seismonitor[key] = val

        df.append(seismonitor)
    df = pd.DataFrame(df)
    df["location"] = df["location"].astype(str)
    # print(df["location"])
    date_cols = ["arrival_time","creation_time",
                "event_start_time","event_end_time",
                "mseed_start_time","mseed_end_time"]
    df[date_cols] = df[date_cols].apply(pd.to_datetime)
    return df

def get_picks(datapicks, datalist, 
             min_p_prob=0.3,
             min_s_prob=0.3,
             one_single_sampling_rate = -1,
            mode='df_obj',
            export=None):   #mode: pick_obj ; df_obj
    """Read phaseNet picks and returns list of Pick objects
    Parameters
    ----------
    datapicks : str
        Path to generated PhaseNet picks csv
    min_prob : float
        Minimum probability to consider a pick.
    
    Returns
    -------
    list
        List of Pick objects
    """

    df = pd.read_csv(datapicks)
    df = df.astype({'itp':'string','tp_prob':'string',
                    'its':'string','ts_prob':'string'})
    df = df.replace("[]",np.nan)
    df = df.dropna(subset=['itp','tp_prob','its','ts_prob'],how="all")

    picks = []
    def _get_picks(irow):
        i,row = irow

        wf_name = row["fname"]
        if not pd.isna(row["itp"]):
            picks_p = row["itp"].strip('[]').strip().split() 
            prob_p = row["tp_prob"].strip('[]').strip().split() 
            P_picks = pick_constructor(datalist,picks_p, 
                                prob_p, wf_name, 'P', min_p_prob,
                                one_single_sampling_rate )
        else:
            P_picks = []
        
        if not pd.isna(row["its"]):
            picks_s = row["its"].strip('[]').strip().split() 
            prob_s = row["ts_prob"].strip('[]').strip().split() 
            S_picks = pick_constructor(datalist,picks_s, 
                            prob_s, wf_name, 'S', min_s_prob,
                            one_single_sampling_rate )
        else:
            S_picks = []

        picks.append(P_picks+S_picks)

    # for irow in df.iterrows():
    #     _get_picks(irow)
    with cf.ThreadPoolExecutor() as executor:
        executor.map(_get_picks,df.iterrows())
    
    picks = [x for n in picks for x in n]

    if mode == 'pick_obj':
        pass
    if (mode ==  'df_obj') or (export != None):
        if not picks:
            return pd.DataFrame()
        picks= picks2df(picks)

        logger = logging.getLogger(f'PhaseNet: picks2df')
        if isinstance(picks, pd.DataFrame) == False:
            logger.warning('There are no picks to convert.')
            return pd.DataFrame()
        else:
            pnet_cols = ["mseed_start_time","mseed_end_time","sampling_rate","sample",
                    "segment","segment_type"]
            picks = picks[SeisMonitor_columns+pnet_cols]
            picks = picks.sort_values('arrival_time')
            picks = picks.reset_index(drop=True)
            logger.info('Conversion pick_sample in pick_time ok.'+
                        ' See your results in pick_df.csv.')
            if export != None:
                picks.to_csv(export,index=False,
                            date_format="%Y-%m-%d %H:%M:%S.%f")
    return picks

def pick_constructor(datalist,picks, prob, wf_name, ph_type, min_prob,
                    one_single_sampling_rate  = -1):
    """Construct Pick objects
    Parameters
    ----------
    picks : list
        List of phase pick sample point in a specific waveform
    prob : list
        List of probabilities associated with the picks
    wf_name : str
        Waveform name
    ph_type : str
        Pick phase type. Can be P or S
    min_prob : str
        Minimum probability to consider a pick
    
    Returns
    -------
    list
        List of picks objects
    """
    ## Add json parameter because need sampling rate of the stations.
    if os.path.isfile(datalist) == False:
        logger = logging.getLogger("Datalist")
        logger.error("Error in phasenet_pick2date()")
        raise Exception("No datalist ")

    else:
        datalist = pd.read_csv(datalist,dtype={"location":str})
        datalist[['mseed_start_time','mseed_end_time']] = datalist[['mseed_start_time','mseed_end_time']].apply(pd.to_datetime)
        datalist = datalist.set_index("fname")


    segment = 0.0
    picks_list = []
    if picks != ['']:
        for pick, prob in zip(picks, prob):
            prob = float(prob)
            if prob >= min_prob:
                

                #----------
                # se obtienen los parámetros para la creación del objeto pick
                #----------
                # algunos datos se obtienen del nombre de la forma de onda
                # net, station, loc, ch, df, *to_segment = wf_name.split('_')
                # datajson[sta]
                wfpath,segment = wf_name.split(".mseed_")
                wfpath = wfpath +".mseed"
                segment = int(segment)

                row = datalist.loc[wfpath]
                net = row["network"]
                sta = row["station"]
                loc = row["location"]
                # print(loc,type(loc))
                ch = row["instrument_type"]
                
                if one_single_sampling_rate == -1:
                    sampling_rate = row["sampling_rate"]
                else:
                    sampling_rate = one_single_sampling_rate

                mseed_starttime = row['mseed_start_time']
                mseed_endtime = row['mseed_end_time']
                sta_lat = row["sta_lat"]
                sta_lon = row["sta_lon"]
                sta_elv = row["sta_elv"]

                dtt = (mseed_endtime - mseed_starttime).seconds

                # se transforma las cuentas asociadas al pick en tiempo
                pick_time, creation_time,obs = sample2time(pick, mseed_starttime, sampling_rate, segment, dtt)
                # se crea el Id usando el tiempo del pick
                ID = id_maker(pick_time, net, sta, loc, ch, ph_type)    

                # Se evalua si la probabilidad es lo suficientemente buena 
                # como para considerarlo manual
                evaluation_mode = 'automatic'
                if prob >= 0.95:
                    evaluation_mode = 'manual'

                # Se crea el objeto Pick
                pick_obj = Pick(resource_id=ResourceIdentifier( id= ID),
                            time=pick_time,
                            time_errors=QuantityError(uncertainty=prob,
                                                    confidence_level=prob*100),
                            waveform_id=WaveformStreamID(network_code=net,
                                                            station_code=sta,
                                                            location_code=loc,
                                                            channel_code=ch,
                                                            resource_uri= ResourceIdentifier( id= ID)),
                            phase_hint = ph_type,
                            evaluation_mode = evaluation_mode,
                            creation_info= CreationInfo(author="PhaseNet_picker",
                                                        creation_time=UTCDateTime.now()),
                            method_id="PhaseNet",
                            comments= [{"event_start_time":None,
                                        "event_end_time":None,
                                        "detection_probability":None,
                                        "snr":None,
                                        "station_lat":sta_lat,
                                        "station_lon":sta_lon,
                                        "station_elv":sta_elv,
                                        "sampling_rate":sampling_rate,
                                        "file_name":wf_name,
                                        "mseed_start_time":mseed_starttime.strftime("%Y-%m-%d %H:%M:%S.%f"),
                                        "mseed_end_time":mseed_endtime.strftime("%Y-%m-%d %H:%M:%S.%f"),
                                        "sample":pick,
                                        "segment":segment,
                                        "segment_type":obs,
                                        }]
                            # comments= [f'{to}to{endtime} sr{df} segment{segment}_sample{pick} {obs}',coords]
                            )
                # Se agrega cada pick a la lista de picks
                picks_list.append(pick_obj)
    return picks_list

def id_maker(pick_time, net, station, loc, ch, phaseHint):
    """Creates the seiscomp Pick PublicID
    
    Parameters
    ----------
    pick_time : Obspy UTCDateTime object
        Time for phase pick
    
    Returns
    ------
    str
       Seiscomp pick PublicID 
    """
    dateID = pick_time.strftime('%Y%m%d.%H%M%S.%f')[:-4]
    if phaseHint == 'P':
        # publicID = dateID+f'-AIC-{net}.{station}.{loc}.{ch}'
        publicID = dateID+f'-P-{net}.{station}.{loc}.{ch}'
    elif phaseHint == 'S':
        # publicID = dateID+f'-S-L2-{net}.{station}.{loc}.{ch}'
        publicID = dateID+f'-S-{net}.{station}.{loc}.{ch}'
    return publicID

def sample2time(sample, to, df, segment, dtt):
    """Transforma las cuentas de un pick de PhaseNet en fecha
    
    Parameters
    ----------
    sample : str
        Sample point of PhaseNet pick
    to : str
        Initial time of the waveform that contains the pick in format
        YYYYmmddHHmmssff. Whit ff as deciseconds
    df : str
        Sampling rate
    
    Returns
    -------
    pick_time : Obspy UTCDateTime object
        Time for phase pick
    creation_time : UTCDateTime
        Time for creation time
    """

    init_time = to
    # if segment is different to 0 which implies that we are using
    # pred_mseed mode
    obs = 'single'
    if segment is not 0:
        
        # if segment is bigger than the lenght of the waveform; then,
        # the segment is an overlapping one, and then we need to
        # include the 1500 samples (15 s) of shiftfing 
        if segment >= dtt*df:
            segment = segment - dtt*df
            init_time -= datetime.timedelta(seconds=1500/df)
            obs = 'overlap'

    add_seconds = (segment + float(sample) )/df
    pick_time = init_time + dt.timedelta( seconds=add_seconds )
    creation_time = UTCDateTime()
    return pick_time, creation_time, obs



# /**
#  * @author [Emmanuel Castillo]
#  * @email [excastillot@unal.edu.co]
#  * @create date 2021-10-12 10:58:25
#  * @modify date 2021-10-12 10:58:25
#  * @desc [description]
#  */
######## 
# Moment magnitude adapted from 
# https://github.com/krischer/moment_magnitude_calculator/tree/master/scripts
# :copyright:
# Lion Krischer (krischer@geophysik.uni-muenchen.de), 2012
from obspy.core.event.resourceid import ResourceIdentifier
from obspy.core.event.base import TimeWindow,WaveformStreamID
from obspy.core.inventory.inventory import Inventory
from obspy.core.stream import Stream
from obspy.io.xseed.parser import Parser
from obspy.core.event import Magnitude as Mag
from obspy.core.event import Catalog
from obspy.core.event import read_events, Comment
from SeisMonitor.utils import isfile
from SeisMonitor.core import utils as scut
from obspy.core.event.base import CreationInfo
import concurrent.futures as cf
from obspy import UTCDateTime
# import mtspec
import numpy as np
import scipy
import os
from SeisMonitor.utils import printlog
from . import utils as ut

class MwPhysicalMagParams():
    def __init__(self,
                vp=4800,
                vsp_factor=1.73,
                density=2650.0,
                waterlevel=10,
                p_radiation_pattern=0.52,
                s_radiation_pattern=0.63):
        self.vp = vp
        self.vsp_factor = vsp_factor
        self.vs = vp/vsp_factor
        self.density = density
        self.waterlevel = waterlevel
        self.p_radiation_pattern = p_radiation_pattern
        self.s_radiation_pattern = s_radiation_pattern

class MwProcessingMagParams():
    def __init__(self,
                time_before_pick = 5,
                time_after_pick = 30,
                padding = 40,
                only_proc_p_pick=False,
                only_proc_s_pick=False):
        self.time_before_pick = time_before_pick
        self.time_after_pick = time_after_pick
        self.padding = padding
        self.only_proc_p_pick = only_proc_p_pick
        self.only_proc_s_pick = only_proc_s_pick

def get_st_according2preference(st,location_list,channel_list):
    """
    Suppose that your preference_type is "location" and your 
    location_list is ["00","20","10"], then this function first
    filter the stream according to location and returns
    a  new stream only with location "00", if no exist "00" will 
    continue with the next preference "20", and otherwise, "10". 
    After that, it is going to take new stream and it will go to 
    filter according to channel_list preference if the new stream 
    has more than one channel type ("HH or BH). 

    Parameters:
    -----------
    st: stream object
        stream 
    location_list: list
        locations in order of the preference ["00","20","10"]
    channel_list: list
        channels in order of the preference ["HH","BH"]

    results:
    --------
    new_st : stream object
        stream according to the preference
    """
    if (st == None) or (len(st)==0) :
        return None

    preference = list(map(lambda x: (x[2],x[3]),st._get_common_channels_info().keys() ))

    if not location_list:
        stats = st[0].stats
        new_st = st
    else:
        locations = list(map(lambda x: x[2],st._get_common_channels_info().keys() ))
    
        index = 0
        loc_pref = None
        # print(len(location_list))
        while index < len(location_list):
            loc_pref = location_list[index]
            if loc_pref in locations:
                index = len(location_list)
            else:
                loc_pref = None
                index += 1

        if loc_pref == None:
            return st

        stats = st[0].stats
        new_st = st.select(network=stats.network, station = stats.station,
                            location=loc_pref)

    if not channel_list:
        pass
    else:
        ## If the same location has two differents sensor
        ## example : 00.HH? or 00.BH?, then choose 
        common_channels = new_st._get_common_channels_info().keys()

        common_channels = list(common_channels)
        common_channels = list(map(lambda x: x[3][:2],common_channels))
        index = 0
        cha_pref = None
        while index < len(channel_list):
            cha_pref = channel_list[index]
            if cha_pref in common_channels:
                index = len(channel_list)
            else:
                cha_pref = None
                index += 1

        if cha_pref == None:
            return st
        else:
            cha_pref = f'{cha_pref}?'
        

        new_st = new_st.select(network=stats.network, station = stats.station,
                                location=loc_pref, channel=cha_pref)

    common_channels = list(map(lambda x: (x[2],x[3]),new_st._get_common_channels_info().keys()))

    printlog("debug",f"Magnitude: {stats.network}-{stats.station}: ",
                f"available:{preference},"+
                f" selected {common_channels} according to preference")
    # logger.info(f"{stats.network}-{stats.station}: "+
    #     f"available:{preference},"+
    #     f" selected {common_channels} according to preference")
    return new_st

class Magnitude():
    def __init__(self,providers,catalog,out_dir) -> None:
        # self.client = client
        # super().__init__(sds_archive,sds_type,format)
        # self.response = response
        self.providers = providers
        print("Reading catalog... ")
        if isinstance(catalog,Catalog):
            self.catalog = catalog
        else:
            self.catalog = read_events(catalog)

        self.agency = self.catalog.creation_info.agency_id
        self.out_dir = out_dir
        self.xml_ml_out_file = os.path.join(out_dir,"Ml_magnitude.xml")    
        self.xml_mw_out_file = os.path.join(out_dir,"Mw_magnitude.xml")    

    def get_Ml(self,
                mag_type="RSNC",
                trimmedtime=50,
                padding=20,waterlevel=10,
                zone=None,
                out_format="QUAKEML"):
        events_mag = []
        for n_ev,event in enumerate(self.catalog,1):
            print(f"Event No. {n_ev}:",event.resource_id,f"from {len(self.catalog)} events")

            if not event.origins:
                print ("No origin for event %s" % event.resource_id)
                continue

            ori_pref  = event.preferred_origin()
            if ori_pref == None:
                event.preferred_origin_id = event.origins[0].resource_id.id
                ori_pref  = event.preferred_origin()

            origin_time = ori_pref.time
            latitude = ori_pref.latitude
            longitude = ori_pref.longitude
            depth = ori_pref.depth
            # depth = ori_pref.depth *1e3

            if latitude ==None or longitude==None:
                continue
            
            amplitudes = []
            station_magnitudes = []
            Mls = []

            def _get_maginfo_by_station(pick):
            # for pick in event.picks:
                
                if pick.phase_hint.upper() != "S":
                    # continue
                    return None
                
                st = self._get_corresponding_st(pick.waveform_id, pick.time,
                                                padding)
                if st == None:
                    # continue
                    return None

                stats = st[0].stats

                ev_params = {"picktime":pick.time, "latitude":latitude,
                            "longitude":longitude}

                resp = Inventory()
                for provider in self.providers:
                    if len(resp) > 0:
                        continue
                    response = provider.inventory
                    # selected_response = response.select(network = pick.waveform_id.network_code,
                    #                                     station = pick.waveform_id.station_code,
                    #                                     location = "*",
                    #                                     channel = pick.waveform_id.channel_code[0:2]+"*")
                    selected_response = response.select(network = stats.network,
                                                        station = stats.station,
                                                        location = stats.location,
                                                        channel = stats.channel[:2]+"*" )
                                    
                    resp = selected_response.__add__(selected_response)

                ampl,epi_dist,tr_id = ut.get_Ml_magparams_by_station(st,resp,
                                ev_params,
                                trimmedtime,
                                waterlevel)
                net,sta,loc,cha = tr_id.split(".")
                amp = ut.write_amplitude_values(ampl,
                                                time_window=TimeWindow(begin=0,
                                                                        end=trimmedtime,
                                                                        reference=ev_params["picktime"]),
                                                pick_id=pick.resource_id,
                                                waveform_id=WaveformStreamID(net,sta,loc,cha),
                                                magnitude_hint="Ml",
                                                agency=self.agency)
                
                amplitudes.append(amp)

                if (ampl == None) or (ampl==0):
                    # continue
                    return None

                
                Ml = ut.get_Ml(ampl,epi_dist,mag_type,zone)
                Mls.append(Ml)

                stamag = ut.write_magsta_values(value=Ml,uncertainty=None,
                                                mag_type="Ml",
                                                origin_id =ori_pref.resource_id,
                                                amplitude_id=amp.resource_id,
                                                method_id=ResourceIdentifier(id="https://docs.obspy.org/tutorial/advanced_exercise/advanced_exercise_solution_3b.html"),
                                                waveform_id=WaveformStreamID(net,sta,loc,cha),
                                                agency=self.agency
                                                 )
                station_magnitudes.append(stamag)

                staname = ".".join((stats.network, stats.station,
                              stats.location ,stats.channel[:2]+"*"))
                print(f"\t-> Ml | {staname}-{pick.phase_hint.upper()} | {Ml}")

            with cf.ThreadPoolExecutor() as executor:
                executor.map(_get_maginfo_by_station,event.picks)
                
            if not Mls:
                Ml = 0
                Ml_std = 0
                print(f"No magnitude for {event.resource_id}")
            else:
                Ml = scipy.stats.trim_mean(Mls,0.25)
                Ml_std = np.array(Mls).std()

            mag = ut.write_magnitude_values(Ml,Ml_std,len(Mls),"Ml",
                           evaluation_mode = "automatic",
                           evaluation_status = "preliminary",
                           method_id=ResourceIdentifier(id="https://docs.obspy.org/tutorial/advanced_exercise/advanced_exercise_solution_3b.html"),
                           origin_id=ori_pref.resource_id,
                           agency=self.agency,
                           comments=None)

            event.amplitudes = amplitudes
            event.station_magnitudes = station_magnitudes
            event.magnitudes.append(mag)
            event.preferred_magnitude_id = mag.resource_id

            events_mag.append(event)
            print(f"{event.resource_id} | Ml: {round(Ml,2)} | Ml_std {round(Ml_std,2)} | stations:{len(Mls)}\n\n")

        # _get_maginfo_by_station(pick)
        catalog = Catalog(events = events_mag )
        catalog = scut.add_aditional_catalog_info(catalog,agency=self.agency)

        if self.xml_ml_out_file != None:
            print (f"Writing output file in {self.xml_ml_out_file}")
            isfile(self.xml_ml_out_file)
            catalog.write(self.xml_ml_out_file, out_format)
        return self.catalog

    def get_Mw(self,physparams,procparams,
            out_format="QUAKEML"):

        Mws = []
        Mws_std = []

        for n_ev,event in enumerate(self.catalog,1):
            print(f"Event No. {n_ev}:",event.resource_id,f"from {len(self.catalog)} events.")

            if not event.origins:
                print ("No origin for event %s" % event.resource_id)
                continue

            origin_time = event.origins[0].time
            latitude = event.origins[0].latitude
            longitude = event.origins[0].longitude
            depth = event.origins[0].depth

            moments = []
            corner_frequencies = []
            for pick in event.picks:

                if procparams.only_proc_s_pick:
                    if pick.phase_hint.upper() == "P":
                        continue

                elif procparams.only_proc_p_pick:
                    if pick.phase_hint.upper() == "S":
                        continue
                    

                st = self._get_corresponding_st(pick.waveform_id, pick.time,
                                                procparams.padding )
                
                if st == None:
                    continue

                resp = Inventory()
                for provider in self.providers:
                    if len(resp) > 0:
                        continue

                    if pick.waveform_id.channel_code == None:
                        channel = "*"
                    else:
                        channel = pick.waveform_id.channel_code[0:2] +"*"

                    response = provider.inventory
                    selected_response = response.select(network = pick.waveform_id.network_code,
                                                        station = pick.waveform_id.station_code,
                                                        location = "*",
                                                        channel = channel )
                                    
                    resp = selected_response.__add__(selected_response)
                st = ut.Mw_st_processing(st, resp,
                                            physparams.waterlevel,
                                             pick.time-procparams.padding)

                if st == None:
                    continue

                traveltime = pick.time - origin_time
                M_0,corner_freqs = ut.get_M0_magnitude_by_pick(st,pick.time,traveltime,
                                                    pick.phase_hint,
                                                    physparams,
                                                    procparams)

                if M_0 != None:
                    staname = ".".join((pick.waveform_id.network_code, pick.waveform_id.station_code,
                              pick.waveform_id.location_code ,channel))
                    
                    Mw_sta = 2.0 / 3.0 * (np.log10(M_0) - 9.1)
                    print(f"\t-> Mw | {staname}-{pick.phase_hint.upper()} | {Mw_sta}")

                    moments.append(M_0)
                    corner_frequencies.extend(corner_freqs)
            
            if not len(moments):
                print (f"No moments could be calculated for event: {event.resource_id}") 
                continue

            # Calculate the seismic moment via basic statistics.
            moments = np.array(moments)
            # moment = moments.mean()
            moment = scipy.stats.trim_mean(moments,0.25)
            moment_std = moments.std()

            corner_frequencies = np.array(corner_frequencies)

            Mw = 2.0 / 3.0 * (np.log10(moment) - 9.1)
            Mw_std = 2.0 / 3.0 * moment_std / (moment * np.log(10))
            Mws_std.append(Mw_std)
            Mws.append(Mw)

            print(f"{event.resource_id} | Mw: {round(Mw,2)} | Mw_std {round(Mw_std,2)} | stations:{len(moments)}\n\n")

            mag = ut.write_magnitude_values(Mw,Mw_std,len(moments),"Mw",
                           evaluation_mode = "automatic",
                           evaluation_status = "preliminary",
                           method_id=ResourceIdentifier("smi:com.github/krischer/moment_magnitude_calculator/automatic/1"),
                           origin_id=event.origins[0].resource_id,
                           agency=self.agency,
                           comments=None)
            event.magnitudes.append(mag)
            event.preferred_magnitude_id = mag.resource_id

        if self.xml_mw_out_file != None:
            print (f"Writing output file in {self.xml_mw_out_file}")
            isfile(self.xml_mw_out_file)
            self.catalog.write(self.xml_mw_out_file, out_format)
        return self.catalog
                
    def _get_corresponding_st(self,waveform_id, 
                                pick_time,
                                padding=20.0,
                                ):
        """
        Helper function to find a requested waveform in the previously created
        waveform_index file.
        Also performs the instrument correction.
        Returns None if the file could not be found.
        """
        
        start = pick_time - padding
        end = pick_time + padding

        if (waveform_id.network_code == None) or\
            (waveform_id.network_code == ""):
            # net = "*"
            net = [ x.waveform_restrictions.network for x in self.providers]
            net = ",".join(net)
        else:
            net = waveform_id.network_code
        if (waveform_id.channel_code == None):
            cha = ""
        else:
            cha = waveform_id.channel_code[0:2]

        stream = Stream()
        for provider in self.providers:

            if len(stream) > 0:
                continue

            client = provider.client
            try:
                st= client.get_waveforms(net,
                                        waveform_id.station_code,
                                        "*",
                                        cha+"*",start,end
                                        )
                st = get_st_according2preference(st,
                                    provider.waveform_restrictions.location_preferences,
                                    provider.waveform_restrictions.channel_preferences)
                stream += st
            except:
                pass

        if len(stream) > 0:
            return st
            
        else:
            print(f"\t-> Not found: {net}-{waveform_id.station_code}"+\
                    f"-*-{cha}* |"+\
                        f"{start} - {end}  ")
            return None
        


if __name__ =="__main__":
    import math
    from obspy.clients.fdsn import Client
    from obspy.core.inventory.inventory import read_inventory
    client = 'http://sismo.sgc.gov.co:8080'
    client = Client(client)

    # catalog = "/home/emmanuel/EDCT/SeisMonitor/data/events/SGC2022knomqj.xml" #3.7
    # catalog = "/home/emmanuel/EDCT/SeisMonitor/data/events/SGC2022cvykgs.xml"
    catalog = "/home/emmanuel/EDCT/SeisMonitor/data/events/SGC2022kszqlt.xml" #2.2 143km
    # catalog = "/home/emmanuel/EDCT/SeisMonitor/data/events/SGC2022ktsrvi.xml" #2 33km
    # catalog = "/home/emmanuel/EDCT/SeisMonitor/data/events/SGC2022krnhiu.xml" #4.3
    # catalog = "/home/emmanuel/EDCT/SeisMonitor/data/events/SGC2022kszqlt.xml" #2.9 103km
    # resp = "/home/emmanuel/Ecopetrol/SeisMonitor/data/metadata/CM.dataless"
    resp = "/home/emmanuel/EDCT/SeisMonitor/data/events/public_CM.xml"
    resp = read_inventory(resp)
    
    out="./test_magnitude,xml"
    mag = Magnitude(client,catalog,resp) 

## prof
    physparams = MwPhysicalMagParams(vp=8200,
                            p_radiation_pattern=0.05,
                            )
    procparams = MwProcessingMagParams(
                            time_before_pick = 2,
                            time_after_pick = 15,
                            only_proc_p_pick=True)

## sup
    # physparams = MwPhysicalMagParams(vp=4800,
    #                         )
    # procparams = MwProcessingMagParams(
    #                         time_before_pick = 0.2,
    #                         time_after_pick = 0.8,
    #                         only_proc_p_pick=True)
    # mag.get_Mw(physparams,procparams,out)

    # ml_params = {"a":1.019,"b":0.0016,"r_ref":140}
    # k = ml_params["a"]*math.log10(ml_params["r_ref"]/100) +\
    #         ml_params["b"]* (ml_params["r_ref"]-100) +3
    # mt = lambda ampl,epi_dist: math.log10(ampl * 1000) + ml_params["a"] * math.log10(epi_dist/ml_params["r_ref"]) +\
    #                     ml_params["b"] * (epi_dist-ml_params["r_ref"]) + k
    mag.get_Ml(mag_type="RSNC")